# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from ...models.tree_ensemble import (
    TreeEnsembleRegressor as _TreeEnsembleRegressor,
    TreeEnsembleClassifier,
)
from ..._deps import _HAS_XGBOOST

import numpy as _np

if _HAS_XGBOOST:
    import xgboost as _xgboost


def recurse_json(
    mlkit_tree,
    xgb_tree_json,
    tree_id,
    node_id,
    feature_map,
    force_32bit_float,
    mode="regressor",
    tree_index=0,
    n_classes=2,
):
    """Traverse through the tree and append to the tree spec.
    """
    relative_hit_rate = None

    try:
        relative_hit_rate = xgb_tree_json["cover"]
    except KeyError:
        pass

    # Fill node attributes
    if "leaf" not in xgb_tree_json:
        branch_mode = "BranchOnValueLessThan"
        split_name = xgb_tree_json["split"]
        feature_index = split_name if not feature_map else feature_map[split_name]

        # xgboost internally uses float32, but the parsing from json pulls it out
        # as a 64bit double.  To trigger the internal float32 detection in the
        # tree ensemble compiler, we need to explicitly cast it to a float 32
        # value, then back to the 64 bit float that protobuf expects.  This is
        # controlled with the force_32bit_float flag.
        feature_value = xgb_tree_json["split_condition"]

        if force_32bit_float:
            feature_value = float(_np.float32(feature_value))

        true_child_id = xgb_tree_json["yes"]
        false_child_id = xgb_tree_json["no"]

        # Get the missing value behavior correct
        missing_value_tracks_true_child = False

        try:
            if xgb_tree_json["missing"] == true_child_id:
                missing_value_tracks_true_child = True
        except KeyError:
            pass

        mlkit_tree.add_branch_node(
            tree_id,
            node_id,
            feature_index,
            feature_value,
            branch_mode,
            true_child_id,
            false_child_id,
            relative_hit_rate=relative_hit_rate,
            missing_value_tracks_true_child=missing_value_tracks_true_child,
        )

    else:
        value = xgb_tree_json["leaf"]
        if force_32bit_float:
            value = float(_np.float32(value))

        if mode == "classifier" and n_classes > 2:
            value = {tree_index: value}

        mlkit_tree.add_leaf_node(
            tree_id, node_id, value, relative_hit_rate=relative_hit_rate
        )

    # Now recurse
    if "children" in xgb_tree_json:
        for child in xgb_tree_json["children"]:
            recurse_json(
                mlkit_tree,
                child,
                tree_id,
                child["nodeid"],
                feature_map,
                force_32bit_float,
                mode=mode,
                tree_index=tree_index,
                n_classes=n_classes,
            )


def convert_tree_ensemble(
    model,
    feature_names,
    target,
    force_32bit_float,
    mode="regressor",
    class_labels=None,
    n_classes=None,
):
    """Convert a generic tree model to the protobuf spec.

    This currently supports:
      * Decision tree regression

    Parameters
    ----------
    model: str | Booster
        Path on disk where the XGboost JSON representation of the model is or
        a handle to the XGboost model.

    feature_names : list of strings or None
        Names of each of the features. When set to None, the feature names are
        extracted from the model.

    target: str,
        Name of the output column.

    force_32bit_float: bool
        If True, then the resulting CoreML model will use 32 bit floats internally.

    mode: str in ['regressor', 'classifier']
        Mode of the tree model.

    class_labels: list[int] or None
        List of classes. When set to None, the class labels are just the range from
        0 to n_classes - 1.

    n_classes: int or None
        Number of classes in classification. When set to None, the number of
        classes is expected from the model or class_labels should be provided.

    Returns
    -------
    model_spec: An object of type Model_pb.
        Protobuf representation of the model
    """
    if not (_HAS_XGBOOST):
        raise RuntimeError("xgboost not found. xgboost conversion API is disabled.")
    accepted_modes = ["regressor", "classifier"]
    if mode not in accepted_modes:
        raise ValueError("mode should be in %s" % accepted_modes)
    import json
    import os

    feature_map = None
    if isinstance(
        model, (_xgboost.core.Booster, _xgboost.XGBRegressor, _xgboost.XGBClassifier)
    ):

        # Testing a few corner cases that we don't support
        if isinstance(model, _xgboost.XGBRegressor):
            if mode == "classifier":
                raise ValueError("mode is classifier but provided a regressor")
            try:
                objective = model.get_xgb_params()["objective"]
            except:
                objective = None
            if objective in ["reg:gamma", "reg:tweedie"]:
                raise ValueError(
                    "Regression objective '%s' not supported for export." % objective
                )

        if isinstance(model, _xgboost.XGBClassifier):
            if mode == "regressor":
                raise ValueError("mode is regressor but provided a classifier")
            n_classes = model.n_classes_
            if class_labels is not None:
                if len(class_labels) != n_classes:
                    raise ValueError(
                        "Number of classes in model (%d) does not match "
                        "length of supplied class list (%d)."
                        % (n_classes, len(class_labels))
                    )
            else:
                class_labels = list(range(n_classes))

        # Now use the booster API.
        if isinstance(model, (_xgboost.XGBRegressor, _xgboost.XGBClassifier)):
            # Name change in 0.7
            if hasattr(model, "get_booster"):
                model = model.get_booster()
            else:
                model = model.booster()

        # Xgboost sometimes has feature names in there. Sometimes does not.
        if (feature_names is None) and (model.feature_names is None):
            raise ValueError(
                "The XGBoost model does not have feature names. They must be provided in convert method."
            )
            feature_names = model.feature_names
        if feature_names is None:
            feature_names = model.feature_names

        xgb_model_str = model.get_dump(with_stats=True, dump_format="json")

        if model.feature_names:
            feature_map = {f: i for i, f in enumerate(model.feature_names)}

    # Path on the file system where the XGboost model exists.
    elif isinstance(model, str):
        if not os.path.exists(model):
            raise TypeError("Invalid path %s." % model)
        with open(model) as f:
            xgb_model_str = json.load(f)

        if feature_names is None:
            raise ValueError(
                "feature names must be provided in convert method if the model is a path on file system."
            )
        else:
            feature_map = {f: i for i, f in enumerate(feature_names)}

    else:
        raise TypeError("Unexpected type. Expecting XGBoost model.")

    if mode == "classifier":
        if n_classes is None and class_labels is None:
            raise ValueError(
                "You must provide class_labels or n_classes when not providing the XGBClassifier"
            )
        elif n_classes is None:
            n_classes = len(class_labels)
        elif class_labels is None:
            class_labels = range(n_classes)
        if n_classes == 2:
            # if we have only 2 classes we only have one sequence of estimators
            base_prediction = [0.0]
        else:
            base_prediction = [0.0 for c in range(n_classes)]
        # target here is the equivalent of output_features in scikit learn
        mlkit_tree = TreeEnsembleClassifier(feature_names, class_labels, target)
        mlkit_tree.set_default_prediction_value(base_prediction)
        if n_classes == 2:
            mlkit_tree.set_post_evaluation_transform("Regression_Logistic")
        else:
            mlkit_tree.set_post_evaluation_transform("Classification_SoftMax")
    else:
        mlkit_tree = _TreeEnsembleRegressor(feature_names, target)
        mlkit_tree.set_default_prediction_value(0.5)

    for xgb_tree_id, xgb_tree_str in enumerate(xgb_model_str):
        if mode == "classifier" and n_classes > 2:
            tree_index = xgb_tree_id % n_classes
        else:
            tree_index = 0

        try:
            # this means that the xgb_tree_str is a json dump and needs to be loaded
            xgb_tree_json = json.loads(xgb_tree_str)
        except:
            # this means that the xgb_tree_str is loaded from a path in file system already and does not need to be reloaded
            xgb_tree_json = xgb_tree_str

        recurse_json(
            mlkit_tree,
            xgb_tree_json,
            xgb_tree_id,
            node_id=0,
            feature_map=feature_map,
            force_32bit_float=force_32bit_float,
            mode=mode,
            tree_index=tree_index,
            n_classes=n_classes,
        )

    return mlkit_tree.spec



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

"""
Neural Network optimization utilities.
"""

import numpy as _np


def _fuse_layer_with_scale_layer(layer_idx, scale_idx, layers):
    layer_type = layers[layer_idx].WhichOneof("layer")
    if layer_type == "convolution":
        layer = layers[layer_idx].convolution
    elif layer_type == "innerProduct":
        layer = layers[layer_idx].innerProduct
    else:
        raise Exception(
            "Scale fusion not supper for layer " "type {} ".format(layer_type)
        )

    scale = layers[scale_idx].scale

    # Update weights
    sw = _np.array(scale.scale.floatValue)
    w = _np.array(layer.weights.floatValue)
    w = w.reshape(layer.outputChannels, int(len(w) / layer.outputChannels))
    wp = w * sw[:, None]
    del layer.weights.floatValue[:]
    layer.weights.floatValue.extend(wp.flatten())

    # Update biases
    if scale.hasBias:
        sb = _np.array(scale.bias.floatValue)
        if not layer.hasBias:
            layer.bias.floatValue.extend(sb)
            layer.hasBias = True
        else:
            lb = _np.array(layer.bias.floatValue)
            bp = sw * lb + sb
            del layer.bias.floatValue[:]
            layer.bias.floatValue.extend(bp)

    # re-wire outputs and delete scale layer
    print("Fused {}->{}".format(layers[layer_idx].name, layers[scale_idx].name))
    del layers[layer_idx].output[:]
    layers[layer_idx].output.extend(layers[scale_idx].output)
    del layers[scale_idx]


def _fuse_layer_with_bias_layer(layer_idx, bias_idx, layers):
    layer_type = layers[layer_idx].WhichOneof("layer")
    if layer_type == "convolution":
        layer = layers[layer_idx].convolution
    elif layer_type == "innerProduct":
        layer = layers[layer_idx].innerProduct
    else:
        raise Exception(
            "Bias fusion not supper for layer " "type {} ".format(layer_type)
        )

    bias = layers[bias_idx].bias

    bb = _np.array(bias.bias.floatValue)
    if not layer.hasBias:
        layer.bias.floatValue.extend(bb)
        layer.hasBias = True
    else:
        lb = _np.array(layer.bias.floatValue)
        bp = lb + bb
        del layer.bias.floatValue[:]
        layer.bias.floatValue.extend(bp)

    # re-wire outputs and delete bias layer
    print("Fused {}->{}".format(layers[layer_idx].name, layers[bias_idx].name))
    del layers[layer_idx].output[:]
    layers[layer_idx].output.extend(layers[bias_idx].output)
    del layers[bias_idx]


def _bn_scale_fusion(bn_idx, scale_idx, layers):
    bn = layers[bn_idx].batchnorm
    scale = layers[scale_idx].scale

    gamma = _np.array(bn.gamma.floatValue)
    beta = _np.array(bn.beta.floatValue)
    sw = _np.array(scale.scale.floatValue)

    gamma = gamma * sw
    beta = beta * sw

    if scale.hasBias:
        sb = _np.array(scale.bias.floatValue)
        beta = beta + sb

    del bn.gamma.floatValue[:]
    del bn.beta.floatValue[:]

    bn.gamma.floatValue.extend(gamma)
    bn.beta.floatValue.extend(beta)

    # re-wire outputs and delete scale layer
    print("Fused {}->{}".format(layers[bn_idx].name, layers[scale_idx].name))
    del layers[bn_idx].output[:]
    layers[bn_idx].output.extend(layers[scale_idx].output)
    del layers[scale_idx]


def _conv_bn_fusion(conv_idx, bn_idx, layers):
    conv = layers[conv_idx].convolution
    bn = layers[bn_idx].batchnorm

    mean = _np.array(bn.mean.floatValue)
    variance = _np.array(bn.variance.floatValue) + bn.epsilon
    gamma = _np.array(bn.gamma.floatValue)
    beta = _np.array(bn.beta.floatValue)
    w = _np.array(conv.weights.floatValue)

    if conv.hasBias:
        b = _np.array(conv.bias.floatValue)
    else:
        b = _np.zeros(conv.outputChannels)

    w = w.reshape(conv.outputChannels, int(len(w) / conv.outputChannels))
    wp = (gamma / _np.sqrt(variance))[:, None] * w
    bp = (gamma * b / _np.sqrt(variance)) - (gamma * mean / _np.sqrt(variance)) + beta

    del conv.weights.floatValue[:]
    if conv.hasBias:
        del conv.bias.floatValue[:]

    conv.weights.floatValue.extend(wp.flatten())
    conv.bias.floatValue.extend(bp)
    conv.hasBias = True

    print("Fused {}->{}".format(layers[conv_idx].name, layers[bn_idx].name))
    # re-wire outputs and delete batchnorm layer
    del layers[conv_idx].output[:]
    layers[conv_idx].output.extend(layers[bn_idx].output)
    del layers[bn_idx]


def _get_nn_mappings(layers):
    layer_map = {}
    type_map = {}
    output_map = {}
    input_map = {}
    for idx, layer in enumerate(layers):
        layer_name = "{}".format(idx)
        layer_map[layer_name] = {"outputs": [], "inputs": []}
        layer_type = layer.WhichOneof("layer")
        if layer_type not in type_map.keys():
            type_map[layer_type] = []
        type_map[layer_type].append(layer_name)

        # Add inputs and outputs for layer
        for o in layer.output:
            layer_map[layer_name]["outputs"].append(o)
        for i in layer.input:
            layer_map[layer_name]["inputs"].append(i)

    # Construct input/output graph dict
    for l in layer_map.keys():
        output_map[l] = []
        input_map[l] = []
        for cl in layer_map.keys():
            if any(x in layer_map[l]["outputs"] for x in layer_map[cl]["inputs"]):
                output_map[l].append(cl)
            if any(x in layer_map[l]["inputs"] for x in layer_map[cl]["outputs"]):
                input_map[l].append(cl)

    return type_map, output_map, input_map


def _optimize_nn(layers):
    type_map, output_map, input_map = _get_nn_mappings(layers)
    bn_layers = []
    conv_layers = []
    ip_layers = []
    bias_layers = []
    scale_layers = []

    # Only fuse with non-instance batchnorm layers
    if "batchnorm" in type_map.keys():
        for bn_layer_idx in type_map["batchnorm"]:
            if not layers[int(bn_layer_idx)].batchnorm.instanceNormalization:
                bn_layers.append(bn_layer_idx)

    if "convolution" in type_map.keys():
        conv_layers = type_map["convolution"]

    if "innerProduct" in type_map.keys():
        ip_layers = type_map["innerProduct"]

    if "bias" in type_map.keys():
        bias_layers = type_map["bias"]

    if "scale" in type_map.keys():
        scale_layers = type_map["scale"]

    # Convolution optimizations
    for conv_idx in conv_layers:
        if len(output_map[conv_idx]) != 1:
            continue
        output_idx = output_map[conv_idx][0]
        if len(input_map[output_idx]) != 1:
            continue

        # Batchnorm fusion
        if output_idx in bn_layers:
            _conv_bn_fusion(int(conv_idx), int(output_idx), layers)
            return _optimize_nn(layers)

        # Scale fusion
        if output_idx in scale_layers:
            _fuse_layer_with_scale_layer(int(conv_idx), int(output_idx), layers)
            return _optimize_nn(layers)

        # Bias fusion
        if output_idx in bias_layers:
            _fuse_layer_with_bias_layer(int(conv_idx), int(output_idx), layers)
            return _optimize_nn(layers)

    # Inner Product optimizations
    for ip_idx in ip_layers:
        if len(output_map[ip_idx]) != 1:
            continue
        output_idx = output_map[ip_idx][0]
        if len(input_map[output_idx]) != 1:
            continue

        # Scale Fusion
        if output_idx in scale_layers:
            _fuse_layer_with_scale_layer(int(ip_idx), int(output_idx), layers)
            return _optimize_nn(layers)

        # Bias Fusion
        if output_idx in bias_layers:
            _fuse_layer_with_bias_layer(int(ip_idx), int(output_idx), layers)
            return _optimize_nn(layers)

    # Batchnorm optimizations
    for bn_idx in bn_layers:
        if len(output_map[bn_idx]) != 1:
            continue
        output_idx = output_map[bn_idx][0]
        if len(input_map[output_idx]) != 1:
            continue

        # Scale Fusion
        if output_idx in scale_layers:
            _bn_scale_fusion(int(bn_idx), int(output_idx), layers)
            return _optimize_nn(layers)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: Model.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import VisionFeaturePrint_pb2 as VisionFeaturePrint__pb2
from . import TextClassifier_pb2 as TextClassifier__pb2
try:
  DataStructures__pb2 = TextClassifier__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = TextClassifier__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = TextClassifier__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = TextClassifier__pb2.FeatureTypes_pb2
from . import WordTagger_pb2 as WordTagger__pb2
try:
  DataStructures__pb2 = WordTagger__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = WordTagger__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = WordTagger__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = WordTagger__pb2.FeatureTypes_pb2
from . import Gazetteer_pb2 as Gazetteer__pb2
try:
  DataStructures__pb2 = Gazetteer__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = Gazetteer__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = Gazetteer__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = Gazetteer__pb2.FeatureTypes_pb2
from . import WordEmbedding_pb2 as WordEmbedding__pb2
try:
  DataStructures__pb2 = WordEmbedding__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = WordEmbedding__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = WordEmbedding__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = WordEmbedding__pb2.FeatureTypes_pb2
from . import ArrayFeatureExtractor_pb2 as ArrayFeatureExtractor__pb2
from . import BayesianProbitRegressor_pb2 as BayesianProbitRegressor__pb2
from . import CategoricalMapping_pb2 as CategoricalMapping__pb2
try:
  DataStructures__pb2 = CategoricalMapping__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = CategoricalMapping__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = CategoricalMapping__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = CategoricalMapping__pb2.FeatureTypes_pb2
from . import CustomModel_pb2 as CustomModel__pb2
from . import DictVectorizer_pb2 as DictVectorizer__pb2
try:
  DataStructures__pb2 = DictVectorizer__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = DictVectorizer__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = DictVectorizer__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = DictVectorizer__pb2.FeatureTypes_pb2
from . import FeatureTypes_pb2 as FeatureTypes__pb2
from . import FeatureVectorizer_pb2 as FeatureVectorizer__pb2
from . import GLMRegressor_pb2 as GLMRegressor__pb2
from . import GLMClassifier_pb2 as GLMClassifier__pb2
try:
  DataStructures__pb2 = GLMClassifier__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = GLMClassifier__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = GLMClassifier__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = GLMClassifier__pb2.FeatureTypes_pb2
from . import NearestNeighbors_pb2 as NearestNeighbors__pb2
try:
  DataStructures__pb2 = NearestNeighbors__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = NearestNeighbors__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = NearestNeighbors__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = NearestNeighbors__pb2.FeatureTypes_pb2
try:
  Parameters__pb2 = NearestNeighbors__pb2.Parameters__pb2
except AttributeError:
  Parameters__pb2 = NearestNeighbors__pb2.Parameters_pb2
try:
  DataStructures__pb2 = NearestNeighbors__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = NearestNeighbors__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = NearestNeighbors__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = NearestNeighbors__pb2.FeatureTypes_pb2
from . import Identity_pb2 as Identity__pb2
from . import Imputer_pb2 as Imputer__pb2
try:
  DataStructures__pb2 = Imputer__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = Imputer__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = Imputer__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = Imputer__pb2.FeatureTypes_pb2
from . import NeuralNetwork_pb2 as NeuralNetwork__pb2
try:
  DataStructures__pb2 = NeuralNetwork__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = NeuralNetwork__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = NeuralNetwork__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = NeuralNetwork__pb2.FeatureTypes_pb2
try:
  Parameters__pb2 = NeuralNetwork__pb2.Parameters__pb2
except AttributeError:
  Parameters__pb2 = NeuralNetwork__pb2.Parameters_pb2
try:
  DataStructures__pb2 = NeuralNetwork__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = NeuralNetwork__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = NeuralNetwork__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = NeuralNetwork__pb2.FeatureTypes_pb2
from . import Normalizer_pb2 as Normalizer__pb2
from . import OneHotEncoder_pb2 as OneHotEncoder__pb2
try:
  DataStructures__pb2 = OneHotEncoder__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = OneHotEncoder__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = OneHotEncoder__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = OneHotEncoder__pb2.FeatureTypes_pb2
from . import Scaler_pb2 as Scaler__pb2
from . import NonMaximumSuppression_pb2 as NonMaximumSuppression__pb2
try:
  DataStructures__pb2 = NonMaximumSuppression__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = NonMaximumSuppression__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = NonMaximumSuppression__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = NonMaximumSuppression__pb2.FeatureTypes_pb2
from . import SVM_pb2 as SVM__pb2
try:
  DataStructures__pb2 = SVM__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = SVM__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = SVM__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = SVM__pb2.FeatureTypes_pb2
from . import TreeEnsemble_pb2 as TreeEnsemble__pb2
try:
  DataStructures__pb2 = TreeEnsemble__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = TreeEnsemble__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = TreeEnsemble__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = TreeEnsemble__pb2.FeatureTypes_pb2
from . import Parameters_pb2 as Parameters__pb2
try:
  DataStructures__pb2 = Parameters__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = Parameters__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = Parameters__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = Parameters__pb2.FeatureTypes_pb2
from . import ItemSimilarityRecommender_pb2 as ItemSimilarityRecommender__pb2
try:
  DataStructures__pb2 = ItemSimilarityRecommender__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = ItemSimilarityRecommender__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = ItemSimilarityRecommender__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = ItemSimilarityRecommender__pb2.FeatureTypes_pb2
from . import SoundAnalysisPreprocessing_pb2 as SoundAnalysisPreprocessing__pb2
from . import LinkedModel_pb2 as LinkedModel__pb2
try:
  Parameters__pb2 = LinkedModel__pb2.Parameters__pb2
except AttributeError:
  Parameters__pb2 = LinkedModel__pb2.Parameters_pb2
try:
  DataStructures__pb2 = LinkedModel__pb2.DataStructures__pb2
except AttributeError:
  DataStructures__pb2 = LinkedModel__pb2.DataStructures_pb2
try:
  FeatureTypes__pb2 = LinkedModel__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = LinkedModel__pb2.FeatureTypes_pb2

from .VisionFeaturePrint_pb2 import *
from .TextClassifier_pb2 import *
from .WordTagger_pb2 import *
from .Gazetteer_pb2 import *
from .WordEmbedding_pb2 import *
from .ArrayFeatureExtractor_pb2 import *
from .BayesianProbitRegressor_pb2 import *
from .CategoricalMapping_pb2 import *
from .CustomModel_pb2 import *
from .DictVectorizer_pb2 import *
from .FeatureTypes_pb2 import *
from .FeatureVectorizer_pb2 import *
from .GLMRegressor_pb2 import *
from .GLMClassifier_pb2 import *
from .NearestNeighbors_pb2 import *
from .Identity_pb2 import *
from .Imputer_pb2 import *
from .NeuralNetwork_pb2 import *
from .Normalizer_pb2 import *
from .OneHotEncoder_pb2 import *
from .Scaler_pb2 import *
from .NonMaximumSuppression_pb2 import *
from .SVM_pb2 import *
from .TreeEnsemble_pb2 import *
from .Parameters_pb2 import *
from .ItemSimilarityRecommender_pb2 import *
from .SoundAnalysisPreprocessing_pb2 import *
from .LinkedModel_pb2 import *

DESCRIPTOR = _descriptor.FileDescriptor(
  name='Model.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x0bModel.proto\x12\x14\x43oreML.Specification\x1a\x18VisionFeaturePrint.proto\x1a\x14TextClassifier.proto\x1a\x10WordTagger.proto\x1a\x0fGazetteer.proto\x1a\x13WordEmbedding.proto\x1a\x1b\x41rrayFeatureExtractor.proto\x1a\x1d\x42\x61yesianProbitRegressor.proto\x1a\x18\x43\x61tegoricalMapping.proto\x1a\x11\x43ustomModel.proto\x1a\x14\x44ictVectorizer.proto\x1a\x12\x46\x65\x61tureTypes.proto\x1a\x17\x46\x65\x61tureVectorizer.proto\x1a\x12GLMRegressor.proto\x1a\x13GLMClassifier.proto\x1a\x16NearestNeighbors.proto\x1a\x0eIdentity.proto\x1a\rImputer.proto\x1a\x13NeuralNetwork.proto\x1a\x10Normalizer.proto\x1a\x13OneHotEncoder.proto\x1a\x0cScaler.proto\x1a\x1bNonMaximumSuppression.proto\x1a\tSVM.proto\x1a\x12TreeEnsemble.proto\x1a\x10Parameters.proto\x1a\x1fItemSimilarityRecommender.proto\x1a SoundAnalysisPreprocessing.proto\x1a\x11LinkedModel.proto\"F\n\x08Pipeline\x12+\n\x06models\x18\x01 \x03(\x0b\x32\x1b.CoreML.Specification.Model\x12\r\n\x05names\x18\x02 \x03(\t\"F\n\x12PipelineClassifier\x12\x30\n\x08pipeline\x18\x01 \x01(\x0b\x32\x1e.CoreML.Specification.Pipeline\"E\n\x11PipelineRegressor\x12\x30\n\x08pipeline\x18\x01 \x01(\x0b\x32\x1e.CoreML.Specification.Pipeline\"m\n\x12\x46\x65\x61tureDescription\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x18\n\x10shortDescription\x18\x02 \x01(\t\x12/\n\x04type\x18\x03 \x01(\x0b\x32!.CoreML.Specification.FeatureType\"\xd6\x01\n\x08Metadata\x12\x18\n\x10shortDescription\x18\x01 \x01(\t\x12\x15\n\rversionString\x18\x02 \x01(\t\x12\x0e\n\x06\x61uthor\x18\x03 \x01(\t\x12\x0f\n\x07license\x18\x04 \x01(\t\x12\x44\n\x0buserDefined\x18\x64 \x03(\x0b\x32/.CoreML.Specification.Metadata.UserDefinedEntry\x1a\x32\n\x10UserDefinedEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xba\x02\n\x10ModelDescription\x12\x37\n\x05input\x18\x01 \x03(\x0b\x32(.CoreML.Specification.FeatureDescription\x12\x38\n\x06output\x18\n \x03(\x0b\x32(.CoreML.Specification.FeatureDescription\x12\x1c\n\x14predictedFeatureName\x18\x0b \x01(\t\x12\"\n\x1apredictedProbabilitiesName\x18\x0c \x01(\t\x12?\n\rtrainingInput\x18\x32 \x03(\x0b\x32(.CoreML.Specification.FeatureDescription\x12\x30\n\x08metadata\x18\x64 \x01(\x0b\x32\x1e.CoreML.Specification.Metadata\"4\n\x0fSerializedModel\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\r\n\x05model\x18\x02 \x01(\x0c\"\x85\x14\n\x05Model\x12\x1c\n\x14specificationVersion\x18\x01 \x01(\x05\x12;\n\x0b\x64\x65scription\x18\x02 \x01(\x0b\x32&.CoreML.Specification.ModelDescription\x12\x13\n\x0bisUpdatable\x18\n \x01(\x08\x12G\n\x12pipelineClassifier\x18\xc8\x01 \x01(\x0b\x32(.CoreML.Specification.PipelineClassifierH\x00\x12\x45\n\x11pipelineRegressor\x18\xc9\x01 \x01(\x0b\x32\'.CoreML.Specification.PipelineRegressorH\x00\x12\x33\n\x08pipeline\x18\xca\x01 \x01(\x0b\x32\x1e.CoreML.Specification.PipelineH\x00\x12;\n\x0cglmRegressor\x18\xac\x02 \x01(\x0b\x32\".CoreML.Specification.GLMRegressorH\x00\x12O\n\x16supportVectorRegressor\x18\xad\x02 \x01(\x0b\x32,.CoreML.Specification.SupportVectorRegressorH\x00\x12M\n\x15treeEnsembleRegressor\x18\xae\x02 \x01(\x0b\x32+.CoreML.Specification.TreeEnsembleRegressorH\x00\x12O\n\x16neuralNetworkRegressor\x18\xaf\x02 \x01(\x0b\x32,.CoreML.Specification.NeuralNetworkRegressorH\x00\x12Q\n\x17\x62\x61yesianProbitRegressor\x18\xb0\x02 \x01(\x0b\x32-.CoreML.Specification.BayesianProbitRegressorH\x00\x12=\n\rglmClassifier\x18\x90\x03 \x01(\x0b\x32#.CoreML.Specification.GLMClassifierH\x00\x12Q\n\x17supportVectorClassifier\x18\x91\x03 \x01(\x0b\x32-.CoreML.Specification.SupportVectorClassifierH\x00\x12O\n\x16treeEnsembleClassifier\x18\x92\x03 \x01(\x0b\x32,.CoreML.Specification.TreeEnsembleClassifierH\x00\x12Q\n\x17neuralNetworkClassifier\x18\x93\x03 \x01(\x0b\x32-.CoreML.Specification.NeuralNetworkClassifierH\x00\x12Y\n\x1bkNearestNeighborsClassifier\x18\x94\x03 \x01(\x0b\x32\x31.CoreML.Specification.KNearestNeighborsClassifierH\x00\x12=\n\rneuralNetwork\x18\xf4\x03 \x01(\x0b\x32#.CoreML.Specification.NeuralNetworkH\x00\x12U\n\x19itemSimilarityRecommender\x18\xf5\x03 \x01(\x0b\x32/.CoreML.Specification.ItemSimilarityRecommenderH\x00\x12\x39\n\x0b\x63ustomModel\x18\xab\x04 \x01(\x0b\x32!.CoreML.Specification.CustomModelH\x00\x12\x39\n\x0blinkedModel\x18\xac\x04 \x01(\x0b\x32!.CoreML.Specification.LinkedModelH\x00\x12=\n\roneHotEncoder\x18\xd8\x04 \x01(\x0b\x32#.CoreML.Specification.OneHotEncoderH\x00\x12\x31\n\x07imputer\x18\xd9\x04 \x01(\x0b\x32\x1d.CoreML.Specification.ImputerH\x00\x12\x45\n\x11\x66\x65\x61tureVectorizer\x18\xda\x04 \x01(\x0b\x32\'.CoreML.Specification.FeatureVectorizerH\x00\x12?\n\x0e\x64ictVectorizer\x18\xdb\x04 \x01(\x0b\x32$.CoreML.Specification.DictVectorizerH\x00\x12/\n\x06scaler\x18\xdc\x04 \x01(\x0b\x32\x1c.CoreML.Specification.ScalerH\x00\x12G\n\x12\x63\x61tegoricalMapping\x18\xde\x04 \x01(\x0b\x32(.CoreML.Specification.CategoricalMappingH\x00\x12\x37\n\nnormalizer\x18\xdf\x04 \x01(\x0b\x32 .CoreML.Specification.NormalizerH\x00\x12M\n\x15\x61rrayFeatureExtractor\x18\xe1\x04 \x01(\x0b\x32+.CoreML.Specification.ArrayFeatureExtractorH\x00\x12M\n\x15nonMaximumSuppression\x18\xe2\x04 \x01(\x0b\x32+.CoreML.Specification.NonMaximumSuppressionH\x00\x12\x33\n\x08identity\x18\x84\x07 \x01(\x0b\x32\x1e.CoreML.Specification.IdentityH\x00\x12L\n\x0etextClassifier\x18\xd0\x0f \x01(\x0b\x32\x31.CoreML.Specification.CoreMLModels.TextClassifierH\x00\x12\x44\n\nwordTagger\x18\xd1\x0f \x01(\x0b\x32-.CoreML.Specification.CoreMLModels.WordTaggerH\x00\x12T\n\x12visionFeaturePrint\x18\xd2\x0f \x01(\x0b\x32\x35.CoreML.Specification.CoreMLModels.VisionFeaturePrintH\x00\x12\x64\n\x1asoundAnalysisPreprocessing\x18\xd3\x0f \x01(\x0b\x32=.CoreML.Specification.CoreMLModels.SoundAnalysisPreprocessingH\x00\x12\x42\n\tgazetteer\x18\xd4\x0f \x01(\x0b\x32,.CoreML.Specification.CoreMLModels.GazetteerH\x00\x12J\n\rwordEmbedding\x18\xd5\x0f \x01(\x0b\x32\x30.CoreML.Specification.CoreMLModels.WordEmbeddingH\x00\x12\x41\n\x0fserializedModel\x18\xb8\x17 \x01(\x0b\x32%.CoreML.Specification.SerializedModelH\x00\x42\x06\n\x04TypeB\x02H\x03P\x00P\x01P\x02P\x03P\x04P\x05P\x06P\x07P\x08P\tP\nP\x0bP\x0cP\rP\x0eP\x0fP\x10P\x11P\x12P\x13P\x14P\x15P\x16P\x17P\x18P\x19P\x1aP\x1b\x62\x06proto3')
  ,
  dependencies=[VisionFeaturePrint__pb2.DESCRIPTOR,TextClassifier__pb2.DESCRIPTOR,WordTagger__pb2.DESCRIPTOR,Gazetteer__pb2.DESCRIPTOR,WordEmbedding__pb2.DESCRIPTOR,ArrayFeatureExtractor__pb2.DESCRIPTOR,BayesianProbitRegressor__pb2.DESCRIPTOR,CategoricalMapping__pb2.DESCRIPTOR,CustomModel__pb2.DESCRIPTOR,DictVectorizer__pb2.DESCRIPTOR,FeatureTypes__pb2.DESCRIPTOR,FeatureVectorizer__pb2.DESCRIPTOR,GLMRegressor__pb2.DESCRIPTOR,GLMClassifier__pb2.DESCRIPTOR,NearestNeighbors__pb2.DESCRIPTOR,Identity__pb2.DESCRIPTOR,Imputer__pb2.DESCRIPTOR,NeuralNetwork__pb2.DESCRIPTOR,Normalizer__pb2.DESCRIPTOR,OneHotEncoder__pb2.DESCRIPTOR,Scaler__pb2.DESCRIPTOR,NonMaximumSuppression__pb2.DESCRIPTOR,SVM__pb2.DESCRIPTOR,TreeEnsemble__pb2.DESCRIPTOR,Parameters__pb2.DESCRIPTOR,ItemSimilarityRecommender__pb2.DESCRIPTOR,SoundAnalysisPreprocessing__pb2.DESCRIPTOR,LinkedModel__pb2.DESCRIPTOR,],
  public_dependencies=[VisionFeaturePrint__pb2.DESCRIPTOR,TextClassifier__pb2.DESCRIPTOR,WordTagger__pb2.DESCRIPTOR,Gazetteer__pb2.DESCRIPTOR,WordEmbedding__pb2.DESCRIPTOR,ArrayFeatureExtractor__pb2.DESCRIPTOR,BayesianProbitRegressor__pb2.DESCRIPTOR,CategoricalMapping__pb2.DESCRIPTOR,CustomModel__pb2.DESCRIPTOR,DictVectorizer__pb2.DESCRIPTOR,FeatureTypes__pb2.DESCRIPTOR,FeatureVectorizer__pb2.DESCRIPTOR,GLMRegressor__pb2.DESCRIPTOR,GLMClassifier__pb2.DESCRIPTOR,NearestNeighbors__pb2.DESCRIPTOR,Identity__pb2.DESCRIPTOR,Imputer__pb2.DESCRIPTOR,NeuralNetwork__pb2.DESCRIPTOR,Normalizer__pb2.DESCRIPTOR,OneHotEncoder__pb2.DESCRIPTOR,Scaler__pb2.DESCRIPTOR,NonMaximumSuppression__pb2.DESCRIPTOR,SVM__pb2.DESCRIPTOR,TreeEnsemble__pb2.DESCRIPTOR,Parameters__pb2.DESCRIPTOR,ItemSimilarityRecommender__pb2.DESCRIPTOR,SoundAnalysisPreprocessing__pb2.DESCRIPTOR,LinkedModel__pb2.DESCRIPTOR,])




_PIPELINE = _descriptor.Descriptor(
  name='Pipeline',
  full_name='CoreML.Specification.Pipeline',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='models', full_name='CoreML.Specification.Pipeline.models', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='names', full_name='CoreML.Specification.Pipeline.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=647,
  serialized_end=717,
)


_PIPELINECLASSIFIER = _descriptor.Descriptor(
  name='PipelineClassifier',
  full_name='CoreML.Specification.PipelineClassifier',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pipeline', full_name='CoreML.Specification.PipelineClassifier.pipeline', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=719,
  serialized_end=789,
)


_PIPELINEREGRESSOR = _descriptor.Descriptor(
  name='PipelineRegressor',
  full_name='CoreML.Specification.PipelineRegressor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pipeline', full_name='CoreML.Specification.PipelineRegressor.pipeline', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=791,
  serialized_end=860,
)


_FEATUREDESCRIPTION = _descriptor.Descriptor(
  name='FeatureDescription',
  full_name='CoreML.Specification.FeatureDescription',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='CoreML.Specification.FeatureDescription.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='shortDescription', full_name='CoreML.Specification.FeatureDescription.shortDescription', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='CoreML.Specification.FeatureDescription.type', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=862,
  serialized_end=971,
)


_METADATA_USERDEFINEDENTRY = _descriptor.Descriptor(
  name='UserDefinedEntry',
  full_name='CoreML.Specification.Metadata.UserDefinedEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='CoreML.Specification.Metadata.UserDefinedEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='CoreML.Specification.Metadata.UserDefinedEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=_descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001')),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1138,
  serialized_end=1188,
)

_METADATA = _descriptor.Descriptor(
  name='Metadata',
  full_name='CoreML.Specification.Metadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='shortDescription', full_name='CoreML.Specification.Metadata.shortDescription', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='versionString', full_name='CoreML.Specification.Metadata.versionString', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='author', full_name='CoreML.Specification.Metadata.author', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='license', full_name='CoreML.Specification.Metadata.license', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='userDefined', full_name='CoreML.Specification.Metadata.userDefined', index=4,
      number=100, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_METADATA_USERDEFINEDENTRY, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=974,
  serialized_end=1188,
)


_MODELDESCRIPTION = _descriptor.Descriptor(
  name='ModelDescription',
  full_name='CoreML.Specification.ModelDescription',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='input', full_name='CoreML.Specification.ModelDescription.input', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='output', full_name='CoreML.Specification.ModelDescription.output', index=1,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='predictedFeatureName', full_name='CoreML.Specification.ModelDescription.predictedFeatureName', index=2,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='predictedProbabilitiesName', full_name='CoreML.Specification.ModelDescription.predictedProbabilitiesName', index=3,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='trainingInput', full_name='CoreML.Specification.ModelDescription.trainingInput', index=4,
      number=50, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='CoreML.Specification.ModelDescription.metadata', index=5,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1191,
  serialized_end=1505,
)


_SERIALIZEDMODEL = _descriptor.Descriptor(
  name='SerializedModel',
  full_name='CoreML.Specification.SerializedModel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='identifier', full_name='CoreML.Specification.SerializedModel.identifier', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='model', full_name='CoreML.Specification.SerializedModel.model', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1507,
  serialized_end=1559,
)


_MODEL = _descriptor.Descriptor(
  name='Model',
  full_name='CoreML.Specification.Model',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='specificationVersion', full_name='CoreML.Specification.Model.specificationVersion', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='description', full_name='CoreML.Specification.Model.description', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='isUpdatable', full_name='CoreML.Specification.Model.isUpdatable', index=2,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pipelineClassifier', full_name='CoreML.Specification.Model.pipelineClassifier', index=3,
      number=200, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pipelineRegressor', full_name='CoreML.Specification.Model.pipelineRegressor', index=4,
      number=201, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pipeline', full_name='CoreML.Specification.Model.pipeline', index=5,
      number=202, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='glmRegressor', full_name='CoreML.Specification.Model.glmRegressor', index=6,
      number=300, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='supportVectorRegressor', full_name='CoreML.Specification.Model.supportVectorRegressor', index=7,
      number=301, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='treeEnsembleRegressor', full_name='CoreML.Specification.Model.treeEnsembleRegressor', index=8,
      number=302, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='neuralNetworkRegressor', full_name='CoreML.Specification.Model.neuralNetworkRegressor', index=9,
      number=303, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='bayesianProbitRegressor', full_name='CoreML.Specification.Model.bayesianProbitRegressor', index=10,
      number=304, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='glmClassifier', full_name='CoreML.Specification.Model.glmClassifier', index=11,
      number=400, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='supportVectorClassifier', full_name='CoreML.Specification.Model.supportVectorClassifier', index=12,
      number=401, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='treeEnsembleClassifier', full_name='CoreML.Specification.Model.treeEnsembleClassifier', index=13,
      number=402, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='neuralNetworkClassifier', full_name='CoreML.Specification.Model.neuralNetworkClassifier', index=14,
      number=403, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='kNearestNeighborsClassifier', full_name='CoreML.Specification.Model.kNearestNeighborsClassifier', index=15,
      number=404, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='neuralNetwork', full_name='CoreML.Specification.Model.neuralNetwork', index=16,
      number=500, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='itemSimilarityRecommender', full_name='CoreML.Specification.Model.itemSimilarityRecommender', index=17,
      number=501, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='customModel', full_name='CoreML.Specification.Model.customModel', index=18,
      number=555, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='linkedModel', full_name='CoreML.Specification.Model.linkedModel', index=19,
      number=556, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='oneHotEncoder', full_name='CoreML.Specification.Model.oneHotEncoder', index=20,
      number=600, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='imputer', full_name='CoreML.Specification.Model.imputer', index=21,
      number=601, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='featureVectorizer', full_name='CoreML.Specification.Model.featureVectorizer', index=22,
      number=602, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dictVectorizer', full_name='CoreML.Specification.Model.dictVectorizer', index=23,
      number=603, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='scaler', full_name='CoreML.Specification.Model.scaler', index=24,
      number=604, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='categoricalMapping', full_name='CoreML.Specification.Model.categoricalMapping', index=25,
      number=606, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='normalizer', full_name='CoreML.Specification.Model.normalizer', index=26,
      number=607, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='arrayFeatureExtractor', full_name='CoreML.Specification.Model.arrayFeatureExtractor', index=27,
      number=609, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='nonMaximumSuppression', full_name='CoreML.Specification.Model.nonMaximumSuppression', index=28,
      number=610, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='identity', full_name='CoreML.Specification.Model.identity', index=29,
      number=900, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='textClassifier', full_name='CoreML.Specification.Model.textClassifier', index=30,
      number=2000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='wordTagger', full_name='CoreML.Specification.Model.wordTagger', index=31,
      number=2001, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='visionFeaturePrint', full_name='CoreML.Specification.Model.visionFeaturePrint', index=32,
      number=2002, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='soundAnalysisPreprocessing', full_name='CoreML.Specification.Model.soundAnalysisPreprocessing', index=33,
      number=2003, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='gazetteer', full_name='CoreML.Specification.Model.gazetteer', index=34,
      number=2004, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='wordEmbedding', full_name='CoreML.Specification.Model.wordEmbedding', index=35,
      number=2005, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='serializedModel', full_name='CoreML.Specification.Model.serializedModel', index=36,
      number=3000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Type', full_name='CoreML.Specification.Model.Type',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=1562,
  serialized_end=4127,
)

_PIPELINE.fields_by_name['models'].message_type = _MODEL
_PIPELINECLASSIFIER.fields_by_name['pipeline'].message_type = _PIPELINE
_PIPELINEREGRESSOR.fields_by_name['pipeline'].message_type = _PIPELINE
_FEATUREDESCRIPTION.fields_by_name['type'].message_type = FeatureTypes__pb2._FEATURETYPE
_METADATA_USERDEFINEDENTRY.containing_type = _METADATA
_METADATA.fields_by_name['userDefined'].message_type = _METADATA_USERDEFINEDENTRY
_MODELDESCRIPTION.fields_by_name['input'].message_type = _FEATUREDESCRIPTION
_MODELDESCRIPTION.fields_by_name['output'].message_type = _FEATUREDESCRIPTION
_MODELDESCRIPTION.fields_by_name['trainingInput'].message_type = _FEATUREDESCRIPTION
_MODELDESCRIPTION.fields_by_name['metadata'].message_type = _METADATA
_MODEL.fields_by_name['description'].message_type = _MODELDESCRIPTION
_MODEL.fields_by_name['pipelineClassifier'].message_type = _PIPELINECLASSIFIER
_MODEL.fields_by_name['pipelineRegressor'].message_type = _PIPELINEREGRESSOR
_MODEL.fields_by_name['pipeline'].message_type = _PIPELINE
_MODEL.fields_by_name['glmRegressor'].message_type = GLMRegressor__pb2._GLMREGRESSOR
_MODEL.fields_by_name['supportVectorRegressor'].message_type = SVM__pb2._SUPPORTVECTORREGRESSOR
_MODEL.fields_by_name['treeEnsembleRegressor'].message_type = TreeEnsemble__pb2._TREEENSEMBLEREGRESSOR
_MODEL.fields_by_name['neuralNetworkRegressor'].message_type = NeuralNetwork__pb2._NEURALNETWORKREGRESSOR
_MODEL.fields_by_name['bayesianProbitRegressor'].message_type = BayesianProbitRegressor__pb2._BAYESIANPROBITREGRESSOR
_MODEL.fields_by_name['glmClassifier'].message_type = GLMClassifier__pb2._GLMCLASSIFIER
_MODEL.fields_by_name['supportVectorClassifier'].message_type = SVM__pb2._SUPPORTVECTORCLASSIFIER
_MODEL.fields_by_name['treeEnsembleClassifier'].message_type = TreeEnsemble__pb2._TREEENSEMBLECLASSIFIER
_MODEL.fields_by_name['neuralNetworkClassifier'].message_type = NeuralNetwork__pb2._NEURALNETWORKCLASSIFIER
_MODEL.fields_by_name['kNearestNeighborsClassifier'].message_type = NearestNeighbors__pb2._KNEARESTNEIGHBORSCLASSIFIER
_MODEL.fields_by_name['neuralNetwork'].message_type = NeuralNetwork__pb2._NEURALNETWORK
_MODEL.fields_by_name['itemSimilarityRecommender'].message_type = ItemSimilarityRecommender__pb2._ITEMSIMILARITYRECOMMENDER
_MODEL.fields_by_name['customModel'].message_type = CustomModel__pb2._CUSTOMMODEL
_MODEL.fields_by_name['linkedModel'].message_type = LinkedModel__pb2._LINKEDMODEL
_MODEL.fields_by_name['oneHotEncoder'].message_type = OneHotEncoder__pb2._ONEHOTENCODER
_MODEL.fields_by_name['imputer'].message_type = Imputer__pb2._IMPUTER
_MODEL.fields_by_name['featureVectorizer'].message_type = FeatureVectorizer__pb2._FEATUREVECTORIZER
_MODEL.fields_by_name['dictVectorizer'].message_type = DictVectorizer__pb2._DICTVECTORIZER
_MODEL.fields_by_name['scaler'].message_type = Scaler__pb2._SCALER
_MODEL.fields_by_name['categoricalMapping'].message_type = CategoricalMapping__pb2._CATEGORICALMAPPING
_MODEL.fields_by_name['normalizer'].message_type = Normalizer__pb2._NORMALIZER
_MODEL.fields_by_name['arrayFeatureExtractor'].message_type = ArrayFeatureExtractor__pb2._ARRAYFEATUREEXTRACTOR
_MODEL.fields_by_name['nonMaximumSuppression'].message_type = NonMaximumSuppression__pb2._NONMAXIMUMSUPPRESSION
_MODEL.fields_by_name['identity'].message_type = Identity__pb2._IDENTITY
_MODEL.fields_by_name['textClassifier'].message_type = TextClassifier__pb2._TEXTCLASSIFIER
_MODEL.fields_by_name['wordTagger'].message_type = WordTagger__pb2._WORDTAGGER
_MODEL.fields_by_name['visionFeaturePrint'].message_type = VisionFeaturePrint__pb2._VISIONFEATUREPRINT
_MODEL.fields_by_name['soundAnalysisPreprocessing'].message_type = SoundAnalysisPreprocessing__pb2._SOUNDANALYSISPREPROCESSING
_MODEL.fields_by_name['gazetteer'].message_type = Gazetteer__pb2._GAZETTEER
_MODEL.fields_by_name['wordEmbedding'].message_type = WordEmbedding__pb2._WORDEMBEDDING
_MODEL.fields_by_name['serializedModel'].message_type = _SERIALIZEDMODEL
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['pipelineClassifier'])
_MODEL.fields_by_name['pipelineClassifier'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['pipelineRegressor'])
_MODEL.fields_by_name['pipelineRegressor'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['pipeline'])
_MODEL.fields_by_name['pipeline'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['glmRegressor'])
_MODEL.fields_by_name['glmRegressor'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['supportVectorRegressor'])
_MODEL.fields_by_name['supportVectorRegressor'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['treeEnsembleRegressor'])
_MODEL.fields_by_name['treeEnsembleRegressor'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['neuralNetworkRegressor'])
_MODEL.fields_by_name['neuralNetworkRegressor'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['bayesianProbitRegressor'])
_MODEL.fields_by_name['bayesianProbitRegressor'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['glmClassifier'])
_MODEL.fields_by_name['glmClassifier'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['supportVectorClassifier'])
_MODEL.fields_by_name['supportVectorClassifier'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['treeEnsembleClassifier'])
_MODEL.fields_by_name['treeEnsembleClassifier'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['neuralNetworkClassifier'])
_MODEL.fields_by_name['neuralNetworkClassifier'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['kNearestNeighborsClassifier'])
_MODEL.fields_by_name['kNearestNeighborsClassifier'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['neuralNetwork'])
_MODEL.fields_by_name['neuralNetwork'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['itemSimilarityRecommender'])
_MODEL.fields_by_name['itemSimilarityRecommender'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['customModel'])
_MODEL.fields_by_name['customModel'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['linkedModel'])
_MODEL.fields_by_name['linkedModel'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['oneHotEncoder'])
_MODEL.fields_by_name['oneHotEncoder'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['imputer'])
_MODEL.fields_by_name['imputer'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['featureVectorizer'])
_MODEL.fields_by_name['featureVectorizer'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['dictVectorizer'])
_MODEL.fields_by_name['dictVectorizer'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['scaler'])
_MODEL.fields_by_name['scaler'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['categoricalMapping'])
_MODEL.fields_by_name['categoricalMapping'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['normalizer'])
_MODEL.fields_by_name['normalizer'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['arrayFeatureExtractor'])
_MODEL.fields_by_name['arrayFeatureExtractor'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['nonMaximumSuppression'])
_MODEL.fields_by_name['nonMaximumSuppression'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['identity'])
_MODEL.fields_by_name['identity'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['textClassifier'])
_MODEL.fields_by_name['textClassifier'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['wordTagger'])
_MODEL.fields_by_name['wordTagger'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['visionFeaturePrint'])
_MODEL.fields_by_name['visionFeaturePrint'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['soundAnalysisPreprocessing'])
_MODEL.fields_by_name['soundAnalysisPreprocessing'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['gazetteer'])
_MODEL.fields_by_name['gazetteer'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['wordEmbedding'])
_MODEL.fields_by_name['wordEmbedding'].containing_oneof = _MODEL.oneofs_by_name['Type']
_MODEL.oneofs_by_name['Type'].fields.append(
  _MODEL.fields_by_name['serializedModel'])
_MODEL.fields_by_name['serializedModel'].containing_oneof = _MODEL.oneofs_by_name['Type']
DESCRIPTOR.message_types_by_name['Pipeline'] = _PIPELINE
DESCRIPTOR.message_types_by_name['PipelineClassifier'] = _PIPELINECLASSIFIER
DESCRIPTOR.message_types_by_name['PipelineRegressor'] = _PIPELINEREGRESSOR
DESCRIPTOR.message_types_by_name['FeatureDescription'] = _FEATUREDESCRIPTION
DESCRIPTOR.message_types_by_name['Metadata'] = _METADATA
DESCRIPTOR.message_types_by_name['ModelDescription'] = _MODELDESCRIPTION
DESCRIPTOR.message_types_by_name['SerializedModel'] = _SERIALIZEDMODEL
DESCRIPTOR.message_types_by_name['Model'] = _MODEL
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Pipeline = _reflection.GeneratedProtocolMessageType('Pipeline', (_message.Message,), dict(
  DESCRIPTOR = _PIPELINE,
  __module__ = 'Model_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.Pipeline)
  ))
_sym_db.RegisterMessage(Pipeline)

PipelineClassifier = _reflection.GeneratedProtocolMessageType('PipelineClassifier', (_message.Message,), dict(
  DESCRIPTOR = _PIPELINECLASSIFIER,
  __module__ = 'Model_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.PipelineClassifier)
  ))
_sym_db.RegisterMessage(PipelineClassifier)

PipelineRegressor = _reflection.GeneratedProtocolMessageType('PipelineRegressor', (_message.Message,), dict(
  DESCRIPTOR = _PIPELINEREGRESSOR,
  __module__ = 'Model_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.PipelineRegressor)
  ))
_sym_db.RegisterMessage(PipelineRegressor)

FeatureDescription = _reflection.GeneratedProtocolMessageType('FeatureDescription', (_message.Message,), dict(
  DESCRIPTOR = _FEATUREDESCRIPTION,
  __module__ = 'Model_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.FeatureDescription)
  ))
_sym_db.RegisterMessage(FeatureDescription)

Metadata = _reflection.GeneratedProtocolMessageType('Metadata', (_message.Message,), dict(

  UserDefinedEntry = _reflection.GeneratedProtocolMessageType('UserDefinedEntry', (_message.Message,), dict(
    DESCRIPTOR = _METADATA_USERDEFINEDENTRY,
    __module__ = 'Model_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.Metadata.UserDefinedEntry)
    ))
  ,
  DESCRIPTOR = _METADATA,
  __module__ = 'Model_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.Metadata)
  ))
_sym_db.RegisterMessage(Metadata)
_sym_db.RegisterMessage(Metadata.UserDefinedEntry)

ModelDescription = _reflection.GeneratedProtocolMessageType('ModelDescription', (_message.Message,), dict(
  DESCRIPTOR = _MODELDESCRIPTION,
  __module__ = 'Model_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.ModelDescription)
  ))
_sym_db.RegisterMessage(ModelDescription)

SerializedModel = _reflection.GeneratedProtocolMessageType('SerializedModel', (_message.Message,), dict(
  DESCRIPTOR = _SERIALIZEDMODEL,
  __module__ = 'Model_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SerializedModel)
  ))
_sym_db.RegisterMessage(SerializedModel)

Model = _reflection.GeneratedProtocolMessageType('Model', (_message.Message,), dict(
  DESCRIPTOR = _MODEL,
  __module__ = 'Model_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.Model)
  ))
_sym_db.RegisterMessage(Model)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
_METADATA_USERDEFINEDENTRY.has_options = True
_METADATA_USERDEFINEDENTRY._options = _descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001'))
# @@protoc_insertion_point(module_scope)



import warnings
import functools


def deprecated(obj=None, suffix=""):
    """
    Decorator to mark a function or a class as deprecated
    """

    def decorator_deprecation_warning(obj):
        @functools.wraps(obj)
        def wrapped(*args, **kwargs):
            if isinstance(obj, type):
                msg = (
                    'Class "%s" is deprecated and will be removed in the next release'
                    % obj.__name__
                )
            else:
                msg = (
                    'Function "%s" is deprecated and will be removed in the next release'
                    % obj.__name__
                )
            if suffix:
                msg += "; %s" % suffix
            warnings.warn(msg, category=FutureWarning)
            return obj(*args, **kwargs)

        return wrapped

    if obj is None:
        return decorator_deprecation_warning

    return decorator_deprecation_warning(obj)



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from ... import SPECIFICATION_VERSION
from ...models._interface_management import set_regressor_interface_params
from ...proto import Model_pb2 as _Model_pb2
from ...proto import FeatureTypes_pb2 as _FeatureTypes_pb2

import numpy as _np

from ..._deps import _HAS_SKLEARN
from ...models import MLModel as _MLModel

if _HAS_SKLEARN:
    from . import _sklearn_util
    import sklearn
    from sklearn.linear_model import LinearRegression

    model_type = "regressor"
    sklearn_class = sklearn.linear_model.LinearRegression


def convert(model, features, target):
    """Convert a linear regression model to the protobuf spec.
    Parameters
    ----------
    model: LinearRegression
        A trained linear regression encoder model.

    feature_names: [str]
        Name of the input columns.

    target: str
        Name of the output column.

    Returns
    -------
    model_spec: An object of type Model_pb.
        Protobuf representation of the model
    """
    if not (_HAS_SKLEARN):
        raise RuntimeError(
            "scikit-learn not found. scikit-learn conversion API is disabled."
        )

    # Check the scikit learn model
    _sklearn_util.check_expected_type(model, LinearRegression)
    _sklearn_util.check_fitted(model, lambda m: hasattr(m, "coef_"))

    return _MLModel(_convert(model, features, target))


def _convert(model, features, target):
    # Set the model class (regressor)
    spec = _Model_pb2.Model()
    spec.specificationVersion = SPECIFICATION_VERSION
    spec = set_regressor_interface_params(spec, features, target)

    # Add parameters for the linear regression.
    lr = spec.glmRegressor

    if isinstance(model.intercept_, _np.ndarray):
        assert len(model.intercept_) == 1
        lr.offset.append(model.intercept_[0])
    else:
        lr.offset.append(model.intercept_)

    weights = lr.weights.add()
    for i in model.coef_:
        weights.value.append(i)
    return spec


def get_input_dimension(model):
    if not (_HAS_SKLEARN):
        raise RuntimeError(
            "scikit-learn not found. scikit-learn conversion API is disabled."
        )
    _sklearn_util.check_fitted(model, lambda m: hasattr(m, "coef_"))
    return model.coef_.size



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import logging
import numpy as np
import copy
import re

from coremltools.converters.mil.mil import Program, Function
from coremltools.converters.mil.mil.passes.pass_registry import PASS_REGISTRY
from coremltools._deps import _IS_MACOS
import PIL.Image


def assert_op_count_match(program, expect, op=None, verbose=False):
    """
    Assert number of ops match expected number. If op is not specified,
    Count total number of ops and match with expect.
    """
    if verbose:
        print(program)

    count = 0
    for _, func in program.functions.items():
        for o in func.operations:
            if not op:
                count += 1
            elif o.op_type.lower() == op.lower():
                count += 1
        np.testing.assert_equal(count, expect)


def assert_model_is_valid(
    program, inputs, backend="nn_proto", verbose=True, expected_output_shapes=None
):
    """
    Assert Core ML model is valid.

    Inputs:

    - input: str -> shape tuple. All program input names need to appear in str.
      shape tuple can only contain positive integers.
    """
    input_dict = dict()
    for name, shape in inputs.items():
        input_dict[name] = np.random.rand(*shape)

    # Avoid circular import
    from coremltools.converters.mil.testing_reqs import ct
    mlmodel = ct.convert(program, source="mil", convert_to=backend)
    assert mlmodel is not None

    if verbose:
        from coremltools.models.neural_network.printer import print_network_spec
        print_network_spec(mlmodel.get_spec(), style="coding")

    if _IS_MACOS:
        prediction = mlmodel.predict(input_dict, useCPUOnly=True)
        assert prediction is not None
        if expected_output_shapes is not None:
            for out_name, out_shape in expected_output_shapes.items():
                assert out_name in prediction
                assert out_shape == prediction[out_name].shape


def assert_same_output_names(prog1, prog2, func_name="main"):
    prog1_outputs = [o.name for o in prog1[func_name].outputs]
    prog2_outputs = [o.name for o in prog2[func_name].outputs]
    assert prog1_outputs == prog2_outputs


def assert_same_output_shapes(prog1, prog2, func_name="main"):
    prog1_output_shapes = [o.shape for o in prog1[func_name].outputs]
    prog2_output_shapes = [o.shape for o in prog2[func_name].outputs]
    assert prog1_output_shapes == prog2_output_shapes


def get_op_types_in_program(prog, func_name="main", skip_const_ops=True):
    """
    Return the operation types in prog[func_name],
    in the same order as they are stored (topological)
    """
    op_types_in_program = []
    for op in prog[func_name].operations:
        if skip_const_ops:
            if op.op_type == "const":
                continue
        op_types_in_program.append(op.op_type)
    return op_types_in_program


def random_gen(
    shape,
    rand_min=0.0,
    rand_max=1.0,
    eps_from_int=0.0,
    allow_duplicate=True,
    dtype=np.float32,
):
    """
    This helper function generates a random array of shape `shape`
    The range of generated numbers will be between (rand_min, rand_max].
    The value of generated numbers will be at least `eps_from_int` apart from integers.
    If allow_duplicate is set to false, it is guaranteed that value generated are all different.
    Default data type is np.float32.
    """
    elem = np.prod(shape).astype(np.int)
    ret = []
    for _ in range(elem):
        while True:
            r = dtype((rand_max - rand_min) * np.random.random() + rand_min)
            if not allow_duplicate and r in ret:
                continue
            if np.issubdtype(dtype, np.integer) or np.fabs(np.round(r) - r) > eps_from_int:
                ret.append(r)
                break
    ret = np.array(ret).reshape(shape)
    return ret.astype(dtype)


def ssa_fn(func):
    """
    Deprecated: use @mb.program()
    """

    def wrapper(*args, **kwargs):
        prog = Program()
        with Function({}) as ssa_func:
            func(*args, **kwargs)

    return wrapper


def to_tuple(v):
    if not isinstance(v, (list, tuple)):
        return tuple([v])
    return tuple(v)


def is_close(expected, actual, atol=1e-04, rtol=1e-05):
    """
    expected, actual: np.array or python primitive (scalar)
    rtol: relative tolerance. See numpy.isclose.
    """

    close = np.isclose(expected, actual, atol=atol, rtol=rtol)
    if not np.all(close):
        diff = expected - actual
        num_not_close = np.sum(~close)
        msg = "Values differ by L1 norm: {}. Num entries not close: {}/{}"
        logging.error(msg.format(np.sum(np.abs(diff)), num_not_close, expected.size))
        if num_not_close < 30:
            logging.error("Differing entries:")
            logging.error("Expected: {}".format(expected[~close]))
            logging.error("Actual: {}".format(actual[~close]))
            logging.error("Delta: {}".format(diff[~close]))
        return False
    return True


def run_core_ml_predict(mlmodel, input_key_values, use_cpu_only=False):
    for k, v in input_key_values.items():
        if isinstance(v, PIL.Image.Image):
            continue
        elif not np.isscalar(v) and not v.shape == ():
            input_key_values[k] = v.astype(np.float32)
        else:
            input_key_values[k] = np.array([v], dtype=np.float32)
    return mlmodel.predict(input_key_values, useCPUOnly=use_cpu_only)

def _get_coreml_out_from_dict(out_dict, out_name):
    if out_name in out_dict:
        return out_dict[out_name]
    elif re.sub("[^a-zA-Z0-9_]", "_", out_name) in out_dict:
        return out_dict[re.sub("[^a-zA-Z0-9_]", "_", out_name)]
    else:
        raise KeyError("{} output not found in Core ML outputs".format(out_name))

def compare_backend(
    mlmodel,
    input_key_values,
    expected_outputs,
    use_cpu_only=False,
    atol=1e-04,
    rtol=1e-05,
    also_compare_shapes=True,
):
    """
    Inputs:
        - mlmodel: MLModel.

        - input_key_values: str -> np.array. Keys must match those in
          input_placeholders.

        - expected_outputs: dict[str, np.array]. Required iff
          frontend_only == False

        - use_cpu_only: True/False.
    """
    if _IS_MACOS:
        pred = run_core_ml_predict(mlmodel, input_key_values,
            use_cpu_only=use_cpu_only)
        if also_compare_shapes:
            compare_shapes(
                mlmodel,
                input_key_values,
                expected_outputs,
                use_cpu_only=use_cpu_only,
                pred=pred,
            )
        if not use_cpu_only:
            atol = max(atol * 100.0, 5e-1)
            rtol = max(rtol * 100.0, 5e-2)
        for o, expected in expected_outputs.items():
            coreml_out = _get_coreml_out_from_dict(pred, o)
            msg = (
                "Output {} differs. useCPUOnly={}.\nInput={}, "
                + "\nExpected={}, \nOutput={}\n"
            )
            assert is_close(expected, coreml_out, atol, rtol), msg.format(
                o, use_cpu_only, input_key_values, expected, coreml_out
            )


def compare_shapes(
    mlmodel, input_key_values, expected_outputs, use_cpu_only=False, pred=None
):
    """
    Inputs:
        - mlmodel: MLModel.

        - input_key_values: str -> np.array or PIL.Image. Keys must match those in
          input_placeholders.

        - expected_outputs: dict[str, np.array].

        - use_cpu_only: True/False.

        - pred: Prediction to use, if it has already been computed.
    """

    if _IS_MACOS:
        if not pred:
            pred = run_core_ml_predict(mlmodel, input_key_values,
                use_cpu_only)
        for o, expected in expected_outputs.items():
            coreml_out = _get_coreml_out_from_dict(pred, o)
            msg = "Output: {}. expected shape {} != actual shape {}".format(
                o, expected.shape, coreml_out.shape
            )
            # Core ML does not support scalar as output
            # remove this special case when support is added
            if expected.shape == () and coreml_out.shape == (1,):
                continue
            assert coreml_out.shape == expected.shape, msg


def get_core_ml_prediction(
    build, input_placeholders, input_values, use_cpu_only=False, 
    backend="nn_proto"):
    """
    Return predictions of the given model.
    """
    program = Program()
    with Function(input_placeholders) as ssa_func:
        output_vars = build(**ssa_func.inputs)
        if isinstance(output_vars, tuple):
            output_vars = list(output_vars)
        elif not isinstance(output_vars, list):
            output_vars = [output_vars]
        ssa_func.set_outputs(output_vars)
        program.add_function("main", ssa_func)

    # Avoid circular import
    from coremltools.converters.mil.testing_reqs import ct
    mlmodel = ct.convert(program, source="mil",
        convert_to=backend, useCPUOnly=use_cpu_only)
    return mlmodel.predict(input_values, useCPUOnly=use_cpu_only)


def apply_pass_and_basic_check(prog, pass_name):
    """
    Apply pass to the program
    """
    prev_prog = copy.deepcopy(prog)
    PASS_REGISTRY[pass_name](prog)
    block = prog.functions["main"]
    prev_block = prev_prog.functions["main"]
    assert_same_output_names(prev_prog, prog)
    assert_same_output_shapes(prev_prog, prog)
    return prev_prog, prev_block, block



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from ._tree_ensemble import convert_tree_ensemble

from ..._deps import _HAS_SKLEARN
from ...models import MLModel as _MLModel

if _HAS_SKLEARN:
    import sklearn.tree as _tree
    from . import _sklearn_util

model_type = "classifier"
sklearn_class = _tree.DecisionTreeClassifier


def convert(model, input_name, output_features):
    """Convert a decision tree model to protobuf format.

    Parameters
    ----------
    decision_tree : DecisionTreeClassifier
        A trained scikit-learn tree model.

    input_name: str
        Name of the input columns.

    output_name: str
        Name of the output columns.

    Returns
    -------
    model_spec: An object of type Model_pb.
        Protobuf representation of the model
    """
    if not (_HAS_SKLEARN):
        raise RuntimeError(
            "scikit-learn not found. scikit-learn conversion API is disabled."
        )

    _sklearn_util.check_expected_type(model, _tree.DecisionTreeClassifier)
    _sklearn_util.check_fitted(
        model, lambda m: hasattr(m, "tree_") and model.tree_ is not None
    )

    return _MLModel(
        convert_tree_ensemble(
            model,
            input_name,
            output_features,
            mode="classifier",
            class_labels=model.classes_,
        )
    )


def supports_output_scores(model):
    return True


def get_output_classes(model):
    return list(model.classes_)


def get_input_dimension(model):
    return model.n_features_



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import unittest
import tempfile
import json

from coremltools._deps import _HAS_SKLEARN, _HAS_XGBOOST

if _HAS_XGBOOST:
    import xgboost
    from coremltools.converters import xgboost as xgb_converter

if _HAS_SKLEARN:
    from sklearn.datasets import load_boston
    from sklearn.ensemble import GradientBoostingRegressor
    from sklearn.preprocessing import OneHotEncoder
    from coremltools.converters import sklearn as skl_converter

# Skipping tests due to <rdar://problem/66545691> xgboost/test_boosted_trees_*.py failing on CI
import pytest
pytestmark = pytest.mark.skip("Skipping for now. See 66545691 for details.")
#

@unittest.skipIf(not _HAS_SKLEARN, "Missing scikit-learn. Skipping tests.")
class GradientBoostingRegressorScikitTest(unittest.TestCase):
    """
    Unit test class for testing scikit-learn converter.
    """

    @classmethod
    def setUpClass(cls):
        """
        Set up the unit test by loading the dataset and training a model.
        """
        if not _HAS_SKLEARN:
            return

        scikit_data = load_boston()
        scikit_model = GradientBoostingRegressor(random_state=1)
        scikit_model.fit(scikit_data["data"], scikit_data["target"])

        # Save the data and the model
        cls.scikit_data = scikit_data
        cls.scikit_model = scikit_model

    def test_conversion(self):
        input_names = self.scikit_data.feature_names
        output_name = "target"
        spec = skl_converter.convert(
            self.scikit_model, input_names, "target"
        ).get_spec()
        self.assertIsNotNone(spec)

        # Test the model class
        self.assertIsNotNone(spec.description)

        # Test the interface class
        self.assertEqual(spec.description.predictedFeatureName, "target")

        # Test the inputs and outputs
        self.assertEqual(len(spec.description.output), 1)
        self.assertEqual(spec.description.output[0].name, "target")
        self.assertEqual(
            spec.description.output[0].type.WhichOneof("Type"), "doubleType"
        )
        for input_type in spec.description.input:
            self.assertEqual(input_type.type.WhichOneof("Type"), "doubleType")
        self.assertEqual(
            sorted(input_names), sorted(map(lambda x: x.name, spec.description.input))
        )

        tr = spec.pipelineRegressor.pipeline.models[
            -1
        ].treeEnsembleRegressor.treeEnsemble
        self.assertIsNotNone(tr)
        self.assertEqual(len(tr.nodes), 1426)

    def test_conversion_bad_inputs(self):

        # Error on converting an untrained model
        with self.assertRaises(Exception):
            model = GradientBoostingRegressor()
            spec = skl_converter.convert(model, "data", "out")

        # Check the expected class during covnersion.
        with self.assertRaises(Exception):
            model = OneHotEncoder()
            spec = skl_converter.convert(model, "data", "out")


@unittest.skipIf(not _HAS_SKLEARN, "Missing scikit-learn. Skipping tests.")
@unittest.skipIf(not _HAS_XGBOOST, "Skipping, no xgboost")
class BoostedTreeRegressorXGboostTest(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        """
        Set up the unit test by loading the dataset and training a model.
        """
        if not _HAS_XGBOOST:
            return
        if not _HAS_SKLEARN:
            return

        scikit_data = load_boston()
        dtrain = xgboost.DMatrix(
            scikit_data.data,
            label=scikit_data.target,
            feature_names=scikit_data.feature_names,
        )
        xgb_model = xgboost.train({}, dtrain, 1)

        # Save the data and the model
        self.scikit_data = scikit_data
        self.xgb_model = xgb_model
        self.feature_names = self.scikit_data.feature_names

    def test_conversion(self):

        feature_names = self.scikit_data.feature_names
        output_name = "target"
        spec = xgb_converter.convert(self.xgb_model, feature_names, "target").get_spec()
        self.assertIsNotNone(spec)

        # Test the model class
        self.assertIsNotNone(spec.description)
        self.assertIsNotNone(spec.treeEnsembleRegressor)

        # Test the interface class
        self.assertEqual(spec.description.predictedFeatureName, "target")

        # Test the inputs and outputs
        self.assertEqual(len(spec.description.output), 1)
        self.assertEqual(spec.description.output[0].name, "target")
        self.assertEqual(
            spec.description.output[0].type.WhichOneof("Type"), "doubleType"
        )
        for input_type in spec.description.input:
            self.assertEqual(input_type.type.WhichOneof("Type"), "doubleType")
        self.assertEqual(
            sorted(self.feature_names),
            sorted(map(lambda x: x.name, spec.description.input)),
        )

        # Test the linear regression parameters.
        tr = spec.treeEnsembleRegressor.treeEnsemble
        self.assertIsNotNone(tr)
        self.assertEqual(len(tr.nodes), 23)

    def test_conversion_from_file(self):

        output_name = "target"
        feature_names = self.feature_names

        xgb_model_json = tempfile.mktemp("tree_model.json")
        xgb_json_out = self.xgb_model.get_dump(dump_format="json")
        with open(xgb_model_json, "w") as f:
            json.dump(xgb_json_out, f)
        spec = xgb_converter.convert(xgb_model_json, feature_names, "target").get_spec()
        self.assertIsNotNone(spec)

        # Test the model class
        self.assertIsNotNone(spec.description)
        self.assertIsNotNone(spec.treeEnsembleRegressor)

        # Test the interface class
        self.assertEqual(spec.description.predictedFeatureName, "target")

        # Test the inputs and outputs
        self.assertEqual(len(spec.description.output), 1)
        self.assertEqual(spec.description.output[0].name, "target")
        self.assertEqual(
            spec.description.output[0].type.WhichOneof("Type"), "doubleType"
        )
        for input_type in spec.description.input:
            self.assertEqual(input_type.type.WhichOneof("Type"), "doubleType")
        self.assertEqual(
            sorted(self.feature_names),
            sorted(map(lambda x: x.name, spec.description.input)),
        )

        # Test the linear regression parameters.
        tr = spec.treeEnsembleRegressor.treeEnsemble
        self.assertIsNotNone(tr)
        self.assertEqual(len(tr.nodes), 23)

    def test_unsupported_conversion(self):

        feature_names = self.scikit_data.feature_names
        output_name = "target"
        xgb_model = xgboost.XGBRegressor(objective="reg:gamma")
        xgb_model.fit(self.scikit_data.data, self.scikit_data.target)
        with self.assertRaises(ValueError):
            spec = xgb_converter.convert(xgb_model, feature_names, "target")

        xgb_model = xgboost.XGBRegressor(objective="reg:tweedie")
        xgb_model.fit(self.scikit_data.data, self.scikit_data.target)
        with self.assertRaises(ValueError):
            spec = xgb_converter.convert(xgb_model, feature_names, "target")

    def test_conversion_bad_inputs(self):

        # Error on converting an untrained model
        with self.assertRaises(TypeError):
            model = GradientBoostingRegressor()
            spec = xgb_converter.convert(model, "data", "out")

        # Check the expected class during conversion
        with self.assertRaises(TypeError):
            model = OneHotEncoder()
            spec = xgb_converter.convert(model, "data", "out")



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import functools
import sympy as sm
from coremltools.converters.mil.mil.types.symbolic import (
    is_symbolic,
    isscalar,
    any_symbolic,
    any_variadic,
)
from coremltools.converters.mil.mil import (
    get_new_symbol,
    get_new_variadic_symbol,
    VALUE,
    SYMBOL,
    types,
)
from ._op_reqs import *


@register_op(doc_str="")
class depth_to_space(Operation):
    """
    Rearrange elements in a tensor from depth (channel) into spatial dimensions.
    
    Parameters
    ----------
    x: tensor<[n, C, H, W], T> (Required)
        * Input tensor of rank ``4``.
    block_size: const i32 (Required)
        * The size of the spatial block. Must be greater than ``1`` and divisible by
          channel dimension ``C``.
    
    Returns
    -------
    tensor<[n, C / block_size^2, H x block_size, W x block_size], T>
        * Where ``b`` is the block size.
    
    Attributes
    ----------
    T: fp32
    """

    input_spec = InputSpec(
        x=TensorInputType(),
        block_size=IntInputType(const=True),
        )

    def __init__(self, **kwargs):
        super(depth_to_space, self).__init__(**kwargs)

    def type_inference(self):
        x_type = self.x.dtype
        n, c, h, w = self.x.shape
        bs = self.block_size.val
        ret_shape = (n, c // (bs * bs), h * bs, w * bs)
        return types.tensor(x_type, ret_shape)


@register_op(doc_str="")
class expand_dims(Operation):
    """
    Insert a single-dimension in a 1-D or higher tensor at each axis in axes.
    
    Parameters
    ----------
    x: tensor<\*?, T> (Required)
        * Scalar or tensor.
    axes: const tensor<[K], i32> Required
        * ``K`` is the number of dimensions expanded.
        * Insert single dimension at dimension index at each axes.
        * Negative value to index from the end. ``-d-1 <= axis <= d``
          where ``d`` is the rank of ``x``.
    
    Returns
    -------
    tensor<\*(rank(x)+K), T>
        * Same type as the input ``x`` with rank ``rank(x)+K``.
    
    Attributes
    ----------
    T: fp32
    """
    
    input_spec = InputSpec(
        x=ScalarOrTensorInputType(), 
        axes=IntTensorInputType(const=True),
    )

    def __init__(self, **kwargs):
        super(expand_dims, self).__init__(**kwargs)

    def type_inference(self):
        x_rank = self.x.rank
        x_type = self.x.dtype
        x_shape = list(self.x.shape)
        axes = self.axes.val
        out_rank = x_rank + len(axes)

        for axis in axes:
            if axis <= -out_rank - 1 or axis >= out_rank:
                msg = 'Axis value {} is out of bounds for {} node "{}" of shape {}'
                raise IndexError(
                    msg.format(axis, self.op_type, self.name, self.x.shape)
                )

        ret_shape = x_shape
        axes = sorted([out_rank + axis if axis < 0 else axis for axis in axes])
        for axis in axes:
            ret_shape.insert(axis, 1)

        return types.tensor(x_type, tuple(ret_shape))

    @precondition(allow=VALUE)
    def value_inference(self):
        axes = self.axes.val
        out_rank = self.x.rank + len(axes)

        for axis in axes:
            if axis <= -out_rank - 1 or axis >= out_rank:
                msg = 'Axis value {} is out of bounds for {} node "{}" of shape {}'
                raise IndexError(
                    msg.format(axis, self.op_type, self.name, self.x.shape)
                )

        axes = sorted([out_rank + axis if axis < 0 else axis for axis in axes])
        ret_shape = list(self.x.shape)
        for axis in axes:
            ret_shape.insert(axis, 1)
        return np.reshape(self.x.val, ret_shape)


def reshape_with_symbol(v, shape):
    """
    Perform basic reshape if v is symbolic (not array of symbols).
    """
    if is_symbolic(v):
        return np.array(v).reshape(shape)
    shape = [int(s) for s in shape]
    return v.reshape(shape)


@register_op(doc_str="")
class reshape(Operation):
    """
    Return a tensor that has the same values as ``x`` with shape ``shape``.
    ``shape`` must have the same volume (number of elements) as ``x``.
    
    Parameters
    ----------
    x: tensor<\*?, T> (Required)
    
        * A n-D tensor or a scalar.
        * If ``x`` is fixed rank (and possibly contains symbolic dimension),
          shape may contain elements that are not positive integers (see below).
        * If ``x`` is variadic rank, shape can only contain positive integers.
        
    shape: tensor<[K], i32> (Required)
    
        A 1-D tensor, with elements from the following:
        
            * Positive integers.
            * Symbols: All but one symbol in shape must be present in ``x.shape``.
              The new symbol that is not present in ``x.shape`` represent a dimension
              such that the total size remains constant. Symbol is illegal
              if ``x`` is variadic rank.
            * ``-1``: ``-1`` introduces a new symbol (see Symbols). Therefore, ``-1`` is
              allowed if all symbols in the shape appear in ``x.shape``. ``-1`` is illegal
              if ``x`` is variadic rank.
            * ``0``: If ``K == rank(x)`` then ``0`` means inheriting from the corresponding
              dimension in ``x.shape``. ``0`` is illegal if ``x`` is variadic rank.
    
    Returns
    -------
    tensor<\*?, T>
        * Tensor with shape determined by the input shape.
    
    Attributes
    ----------
    T: fp32
    """

    input_spec = InputSpec(
        x=ScalarOrTensorInputType(),
        shape=IntTensorInputType(),
        )

    def __init__(self, **kwargs):
        super(reshape, self).__init__(**kwargs)

    def type_inference(self):
        if any_symbolic(self.shape.shape):
            # We can't infer any shape if shape has variable length.
            return types.tensor(self.x.dtype, (get_new_variadic_symbol(),))

        # shape has fixed length here.
        if self.shape.sym_val is None:
            shape = tuple([get_new_symbol() for _ in range(self.shape.shape[0])])
            return types.tensor(self.x.dtype, shape)
        t, _ = self._get_type_val()
        return t

    @precondition(allow=VALUE | SYMBOL)
    def value_inference(self):
        _, val = self._get_type_val()
        return val

    def _get_type_val(self):
        x_type = self.x.dtype
        x_shape = self.x.shape
        x_vol = np.prod(x_shape)
        # shape is const, and thus sym_val is not None
        sym_shape = self.shape.sym_val
        sym_shape = [get_new_symbol() if d == -1 else d for d in sym_shape]
        try:
            ret_shape = reshape.enforce_volumetric_constraint(x_vol, sym_shape)
        except:
            ret_shape = sym_shape
        ret_val = None
        if self.x.val is not None and all(isscalar(a) and not is_symbolic(a) for a in ret_shape):
            ret_val = reshape_with_symbol(self.x.val, ret_shape)
        return types.tensor(x_type, tuple(ret_shape)), ret_val

    @staticmethod
    def enforce_volumetric_constraint(left_volume, inshape):
        left_symbols = set()
        if is_symbolic(left_volume):
            left_symbols = left_volume.free_symbols
        # Generally, we want to solve for right in terms of left. But this
        # is kinda annoying actually.
        shape = list(inshape)

        # Handling when reshape is given 0 instead of actual input
        # input tensor shape: [4, 3, 2], reshape:[0, -1], output tensor shape: [4, 6]
        if shape.count(-1) > 1:
            raise ValueError(
                "Reshape op supports only one dimension to be -1. Given {}".format(
                    shape.count(-1)
                )
            )

        infer_dim_index = shape.index(-1) if -1 in shape else None
        right_volume = 1
        for i in shape:
            if i != -1:
                right_volume = right_volume * i

        if infer_dim_index:
            shape[infer_dim_index] = left_volume // right_volume

        if not is_symbolic(right_volume):
            return shape

        constraints = [left_volume - right_volume]
        solve_for = [s for s in shape if is_symbolic(s)]

        for rightsym in solve_for:
            sol = sm.solve(constraints, [rightsym], dict=True)
            if not isinstance(sol, list):
                sol = [sol]
            # look for an acceptable solution
            for s in sol:
                if 0 in s.values():
                    continue
                for i in range(len(shape)):
                    if shape[i] in s:
                        v = s[shape[i]]
                        if len(v.free_symbols - left_symbols) > 0:
                            continue
                        try:
                            shape[i] = int(v)
                        except:
                            shape[i] = v
        return shape


@register_op(doc_str="")
class reverse(Operation):
    """
    Reverse the order of the input tensor ``x`` along specified ``axes``(dimensions).
    
    Parameters
    ----------
    x: tensor<\*?, T> (Required)
        * Input tensor.
    
    axes: const<D, i32> (Optional)
        * Dimension(s) to reverse. Each axis must be in the range ``[-rank(x), rank(x))``.
        * Defaults to None (reverse on all dimensions).
    
    Returns
    -------
    tensor<\*?, T>
        * Same type and shape as the input tensor.
    
    Attributes
    ----------
    T: fp32
    
    References
    ----------
    See `tf.reverse <https://www.tensorflow.org/api_docs/python/tf/reverse>`_
    and `TORCH <https://pytorch.org/docs/stable/torch.html#torch.flip>`_.
    """
    
    input_spec = InputSpec(
        x=TensorInputType(),
        axes=IntTensorInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(
            axes=None,
            )

    def __init__(self, **kwargs):
        super(reverse, self).__init__(**kwargs)

    def type_inference(self):
        return self.x.sym_type

    @precondition(allow=VALUE)
    def value_inference(self):
        res = self.x.val
        axes = self.axes.val if self.axes is not None else range(self.x.rank)
        for axis in axes:
            res = np.flip(res, axis=axis)
        return res


@register_op(doc_str="")
class reverse_sequence(Operation):
    """
    Reverse variable length slices for specified axes / dimensions of the input
    tensor. This op first slices input tensor along the ``batch_axis`` dimension, then
    partially reverses the elements along the ``seq_axis`` for the first ``lengths[i]``
    elements.

    Parameters
    ----------
    x: tensor<\*?, T> (Required)
        * Input tensor.
    lengths: tensor<L, i32> (Required)
        * 1-dimensional tensor of length ``x.shape[batch_axis]`` specifying the length
          of the sequence to reverse.
        * Values must be in range ``[0, x.shape[seq_axis]]``.
    seq_axis: const<i32> (Optional)
        * The dimension to reverse.
        * Defaults to ``0``.
    batch_axis: const<i32> (Optional)
        * Dimension for slicing.
        * Defaults to ``0``.
    
    Returns
    -------
    tensor<\*?, T>
        * Same type and shape as the input tensor.
    
    Attributes
    ----------
    T: fp32
    
    References
    ----------
    `tf.reverse_sequence <https://www.tensorflow.org/api_docs/python/tf/reverse_sequence>`_
    
    """
    
    input_spec = InputSpec(
        x=TensorInputType(),
        lengths=IntTensorInputType(),
        seq_axis=IntInputType(const=True, optional=True),
        batch_axis=IntInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(
            seq_axis=0,
            batch_axis=0)

    def __init__(self, **kwargs):
        super(reverse_sequence, self).__init__(**kwargs)

    def type_inference(self):
        return self.x.sym_type

    @precondition(allow=VALUE)
    def value_inference(self):
        raise NotImplementedError("TODO")


@register_op(doc_str="")
class slice_by_index(Operation):
    """
    Method for numpy style indexing and slicing.
    Suppose we have a tensor ``x``, this method achieves:
    ``result = x[begin[0]: end[0]: stride[0], begin[1]: end[1]: stride[1], ...]``
    Note this method does not support pure indexing. You would need to do squeeze if indexing is intended.

    Parameters
    ----------
    x: tensor<*?, T> (Required)
        * Input tensor
    begin: tensor<[rank<x>], i32> (Required)
        * Starting index for the dimension of slicing.
    end: tensor<[rank(x)], i32> (Required)
        * Ending index for the dimension of slicing.
    stride: tensor<[rank(x)], i32> (Optional)
        * Default as all ``1``s.
        * Stride for the dimension of slicing.
    begin_mask: tensor<[rank(x)], bool> (Optional)
        * Default to all ``False``.
        * If ``begin_mask[i]==True``, neglect ``begin[i]``, and set ``begin[i]`` to ``0``.
    end_mask: tensor<[rank(x)], bool> (Optional)
        * Default to all ``False``.
        * If ``end_mask[i]==True``, neglect ``end[i]``, and set ``end[i]`` to ``x.shape[i]``.
    squeeze_mask: tensor<[rank(x)], bool> (Optional)
        * Default to all ``False``.
        * If ``squeeze_mask[i]==true``, neglect ``end[i]``, and do the pure index at ``begin[i]``.

    Returns
    -------
    tensor<*?, T>
        - Scalar or tensor.

    Attributes
    ----------
    T: fp32

    """

    input_spec = InputSpec(
        x=TensorInputType(),
        begin=IntTensorInputType(),
        end=IntTensorInputType(),
        stride=IntTensorInputType(const=True, optional=True),
        begin_mask=BoolTensorInputType(const=True, optional=True),
        end_mask=BoolTensorInputType(const=True, optional=True),
        squeeze_mask=BoolTensorInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(
            stride=None,
            begin_mask=None,
            end_mask=None,
            squeeze_mask=None,
            )

    def __init__(self, **kwargs):
        super(slice_by_index, self).__init__(**kwargs)

    def type_inference(self):

        # get tensor and set default value
        begin = self.begin.val
        end = self.end.val
        x_rank = self.x.rank
        stride = self.stride.val if self.stride is not None else [1] * x_rank
        begin_mask = (
            self.begin_mask.val if self.begin_mask is not None else [False] * x_rank
        )
        end_mask = self.end_mask.val if self.end_mask is not None else [False] * x_rank
        squeeze_mask = (
            self.squeeze_mask.val if self.squeeze_mask is not None else [False] * x_rank
        )

        # solve shape
        x_shape = self.x.shape
        ret_shape = []

        if begin is None or len(begin) == 0:
            begin = [None] * len(x_shape)
        if end is None or len(end) == 0:
            end = [None] * len(x_shape)

        # solve for shape inference
        for idx in range(len(x_shape)):
            # skip if we want to squeeze the dimension
            if squeeze_mask[idx]:
                continue

            # for those a[:] cases
            if begin_mask[idx] and end_mask[idx]:
                if is_symbolic(x_shape[idx]):
                    if stride[idx] == -1 or stride[idx] == 1:
                        ret_shape.append(x_shape[idx])
                    else:
                        ret_shape.append(get_new_symbol())
                    continue
                else:
                    num = np.ceil(float(x_shape[idx]) / abs(stride[idx])).astype(
                        np.int32
                    )
                    ret_shape.append(num)
                    continue

            # for symbolic case
            if is_symbolic(x_shape[idx]):
                ret_shape.append(get_new_symbol())
                continue

            # when begin and end are not determined
            if begin[idx] is None and not begin_mask[idx]:
                ret_shape.append(get_new_symbol())
                continue
            if end[idx] is None and not end_mask[idx]:
                ret_shape.append(get_new_symbol())
                continue

            # parse negative dimention
            if begin[idx] is not None and begin[idx] < 0:
                begin[idx] = max(0, begin[idx] + x_shape[idx])
            if end[idx] is not None and end[idx] < 0:
                end[idx] = max(0, end[idx] + x_shape[idx])

            # compute shape
            low, high = [0, x_shape[idx]] if stride[idx] > 0 else [-1, x_shape[idx] - 1]
            begin_idx, end_idx = (
                [begin[idx], end[idx]] if stride[idx] > 0 else [end[idx], begin[idx]]
            )
            is_begin_mask, is_end_mask = (
                [begin_mask[idx], end_mask[idx]]
                if stride[idx] > 0
                else [end_mask[idx], begin_mask[idx]]
            )
            if is_begin_mask:
                begin_idx = low
            end_idx = high if is_end_mask else min(end_idx, high)
            num = np.ceil(float(end_idx - begin_idx) / abs(stride[idx])).astype(
                np.int32
            )
            ret_shape.append(max(0.0, num))

        if len(ret_shape) == 0:
            # Scalar case.
            return self.x.dtype
        else:
            return types.tensor(self.x.dtype, tuple(ret_shape))

    def value_inference(self):
        if self.x.sym_val is None or self.begin.val is None or self.end.val is None:
            return None
        x_shape = self.x.shape
        begin = [int(i) for i in list(self.begin.val[:])]
        end = [int(i) for i in list(self.end.val[:])]
        stride = [1] * self.x.rank if self.stride is None else self.stride.val
        begin_mask = (
            [False] * self.x.rank if self.begin_mask is None else self.begin_mask.val
        )
        end_mask = [False] * self.x.rank if self.end_mask is None else self.end_mask.val
        squeeze_mask = (
            [False] * self.x.rank
            if self.squeeze_mask is None
            else self.squeeze_mask.val
        )

        slices = []
        for idx, mask in enumerate(begin_mask):
            if mask:
                begin[idx] = None
        for idx, mask in enumerate(end_mask):
            if mask:
                end[idx] = None
        squeeze_axes = []
        for idx, mask in enumerate(squeeze_mask):
            if mask:
                end[idx] = None
                stride[
                    idx
                ] = 2147483647  # We slice out only 1 element by setting stride to INF
                squeeze_axes.append(idx)
        for idx in range(self.x.rank):
            slices.append(slice(begin[idx], end[idx], stride[idx]))

        slices = tuple(slices)
        res = self.x.sym_val[slices]

        # remove squeezed axes
        if len(squeeze_axes) > 0:
            if len(squeeze_axes) == len(res.shape):
                if len(res) == 0:
                    logging.warning("%s seems to be a 0 sized tensor", self.name)
                    return np.array([])
                res = res.tolist()[0]
                if is_symbolic(res):
                    return res
                elif self.x.dtype == types.int32 or self.x.dtype == types.int64:
                    res = np.int32(res)
                elif self.x.dtype == types.float or self.x.dtype == types.double:
                    res = np.float32(res)
                else:
                    raise ValueError(
                        "Unable to convert type {}".format(self.x.sym_val.dtype)
                    )
            else:
                res = np.squeeze(res, axis=tuple(squeeze_axes))
        return res


@register_op(doc_str="")
class slice_by_size(Operation):
    """
    Perform numpy-style indexing and slicing. For example, if you have a
    tensor ``x``, this method produces:

    ``result = x[begin[0]: end[0]: stride[0], begin[1]: end[1]: stride[1], ...]``

    Note: This method does not support pure indexing. You would need to do a squeeze if
    indexing is intended.

    Parameters
    ----------
    x: tensor<\*?, T> (Required)
        * Input tensor.

    begin: tensor<[rank<x>], i32> (Required)
        * Starting index for the dimension of slicing.

    end: tensor<[rank(x)], i32> (Required)
        * Ending index for the dimension of slicing.

    stride: tensor<[rank(x)], i32> (Optional)
        * Default to all ones (``1``).
        * Stride for the dimension of slicing.

    begin_mask: tensor<[rank(x)], bool> (Optional)
        * Default to all ``False``.
        * If ``begin_mask[i]==True``, neglect ``begin[i]``, and set ``begin[i]`` to ``0``.

    end_mask: tensor<[rank(x)], bool> (Optional)
        * Default to all ``False``.
        * If ``end_mask[i]==True``, neglect ``end[i]``, and set ``end[i]``
          to ``x.shape[i]``.

    squeeze_mask: tensor<[rank(x)], bool> (Optional)
        * Default to all ``False``.
        * If ``squeeze_mask[i]==true``, neglect ``end[i]``, and do the pure index at
          ``begin[i]``.

    Returns
    -------
    tensor<\*?, T>
        * Scalar or tensor.

    Attributes
    ----------
    T: fp32
    """

    input_spec = InputSpec(
        x=TensorInputType(),
        begin=IntTensorInputType(),
        size=IntTensorInputType(),
    )

    def __init__(self, **kwargs):
        super(slice_by_size, self).__init__(**kwargs)

    def type_inference(self):
        if self.begin.rank != 1:
            raise ValueError(
                "begin should be 1-D tensor, got {}-D tensor instead".format(
                    self.begin.rank
                )
            )
        if self.size.rank != 1:
            raise ValueError(
                "size should be 1-D tensor, got {}-D tensor instead".format(
                    self.size.rank
                )
            )
        if self.x.rank != self.begin.shape[0]:
            raise ValueError(
                "Length of begin {} doesn't equal to input rank {}.".format(
                    len(self.begin.shape[0]), len(self.x.rank)
                )
            )
        if self.x.rank != self.size.shape[0]:
            raise ValueError(
                "Length of size {} doesn't equal to input rank {}.".format(
                    len(self.size.shape[0]), len(self.x.rank)
                )
            )

        x_shape = self.x.shape
        ret_shape = []
        if self.size.sym_val is None:
            ret_shape = [get_new_symbol() for _ in range(self.x.rank)]
            return types.tensor(self.x.dtype, tuple(ret_shape))

        for idx, s in enumerate(self.size.sym_val):
            if is_symbolic(s):
                ret_shape.append(s)
            elif s != -1:
                ret_shape.append(s)
            elif self.begin.sym_val is not None:
                ret_shape.append(x_shape[idx] - self.begin.sym_val[idx])
            else:
                ret_shape.append(get_new_symbol())

        return types.tensor(self.x.dtype, tuple(ret_shape))

    @precondition(allow=VALUE | SYMBOL)
    def value_inference(self):
        if any_symbolic(self.begin.sym_val):
            return None
        if any_symbolic(self.size.sym_val):
            return None
        if self.x.val is None:
            return None
        slices = []
        for i in range(self.x.rank):
            begin_val = self.begin.val[i]
            if begin_val < 0:
                if is_symbolic(self.x.shape[i]):
                    return None
                begin_val += self.x.shape[i]
            if self.size.val[i] > 0:
                slices.append(slice(begin_val, begin_val + self.size.val[i]))
            else:
                slices.append(slice(begin_val, None, None))
        return self.x.val[tuple(slices)]


@register_op(doc_str="")
class space_to_depth(Operation):
    """
    Rearrange elements in a tensor from spatial into depth (channel) dimension.
    
    Parameters
    ----------
    x: tensor<[n, C, H, W], T> (Required)
        * Input tensor of rank ``4``.
    block_size: const<i32> (Required)
        * The size of the spatial block. Must be greater than ``1`` and divisible
          by spatial dimensions ``H, W``.
    
    Returns
    -------
    tensor<[n, C x block_size^2, H / block_size, W / block_size], T>
        * Where ``b`` is the block size.
    
    Attributes
    ----------
    T: fp32
    """

    input_spec = InputSpec(
        x=TensorInputType(),
        block_size=IntInputType(const=True),)

    def __init__(self, **kwargs):
        super(space_to_depth, self).__init__(**kwargs)

    def type_inference(self):
        x_type = self.x.dtype
        n, c, h, w = self.x.shape
        bs = self.block_size.val
        ret_shape = (n, c * (bs * bs), h // bs, w // bs)
        return types.tensor(x_type, ret_shape)


@register_op(doc_str="")
class squeeze(Operation):
    """
    Remove single-dimension dimensions in a 1-D or higher tensor.
    
    Parameters
    ----------
    x: tensor<\*?,T> (Required)
        * Must be at least 1-D.
    axes: const<K,i32> (Optional)
        * Axes to squeeze out.
        * Default to remove all single-dimensions.
    
    Returns
    -------
    tensor<\*(rank(x)-K),T>
        * Tensor with same type as input ``x`` and rank ``rank(x)-K``.
    
    Attributes
    ----------
    T: fp32
    """
    
    input_spec = InputSpec(
        x=TensorInputType(),
        axes=IntTensorInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(
            axes=None,
            )

    def __init__(self, **kwargs):
        super(squeeze, self).__init__(**kwargs)

    def type_inference(self):
        x_type = self.x.dtype
        x_shape = self.x.shape
        squeezed_shape = list(x_shape)
        if self.axes is None:
            # Squeeze all single-dim, assuming symbolic dims != 1
            squeezed_shape = [s for s in squeezed_shape if s != 1]
        else:
            axes = self.axes.val
            axes = [axis if axis >= 0 else axis + self.x.rank for axis in axes]
            for i in sorted(axes)[::-1]:  # descending order
                if len(squeezed_shape) <= i:
                    raise ValueError(
                        "Cannot squeeze dim {} for shape"
                        + " {}".format(i, squeezed_shape)
                    )
                squeezed_shape.pop(i)

        return types.tensor(x_type, tuple(squeezed_shape))

    @precondition(allow=VALUE)
    def value_inference(self):
        if self.x.val is None:
            return None
        if self.axes is None:
            return np.squeeze(self.x.val)
        else:
            return np.squeeze(self.x.val, axis=tuple(self.axes.val))


@register_op(doc_str="")
class transpose(Operation):
    """
    Permute tensor ``x`` dimensions according to ``perm``.
    
    Parameters
    ----------
    x: tensor<\*?, T> (Required)
        * Must be at least 1-D. ``x`` may have a symbolic shape.
    perm: const<[rank(x)], i32> (Required)
        * Permutation order. Must be non-negative integers.
    
    Returns
    -------
    tensor<\*?,T>
        * Tensor with same rank and type as ``x``.
    
    Attributes
    ----------
    T: fp32
    
    References
    ----------
    `torch.Tensor.permute <https://pytorch.org/docs/stable/tensors.html#torch.Tensor.permute>`_
    """

    input_spec = InputSpec(
        x=TensorInputType(),
        perm=IntTensorInputType(const=True),)

    def __init__(self, **kwargs):
        super(transpose, self).__init__(**kwargs)

    def type_inference(self):
        x_type = self.x.dtype
        perm = self.perm.val
        x_shape = np.array(self.x.shape)
        if len(perm) != self.x.rank:
            msg = "perm should have the same length as rank(x): {} != {}"
            raise ValueError(msg.format(len(perm), self.x.rank))
        if self.x.rank == 0:
            return self.x.sym_type  # scalar cannot be transposed
        if any_variadic(self.x.shape):
            ret_shape = get_new_variadic_symbol()
        else:
            ret_shape = x_shape[perm]
        return types.tensor(x_type, tuple(ret_shape))

    @precondition(allow=VALUE | SYMBOL)
    def value_inference(self):
        return np.transpose(self.x.val, axes=self.perm.val)


@register_op(doc_str="")
class pixel_shuffle(Operation):
    """
    Rearrange elements in a tensor from depth (channel) into spatial dimensions.
    Equivalent to PyTorch's ``PixelShuffle``.
    
    Parameters
    ----------
    x: tensor<[n, C x f^2, H, W], T> (Required)
        * Input tensor of rank ``4``.
    upscale_factor: const<i32>
        * Factor to increase spatial resolution by.
    
    Returns
    -------
    tensor<[n, C, H x f, W x f], T>
        * Where ``f`` is the upscale factor.
    
    Attributes
    ----------
    T: fp32
    
    References
    ----------
    `torch.nn.PixelShuffle <https://pytorch.org/docs/stable/generated/torch.nn.PixelShuffle.html?highlight=pixel%20shuffle#torch.nn.PixelShuffle>`_
    """
    
    input_spec = InputSpec(
        x=TensorInputType(), upscale_factor=IntInputType(const=True),
    )

    def __init__(self, **kwargs):
        super(pixel_shuffle, self).__init__(**kwargs)

    def type_inference(self):
        x_type = self.x.dtype
        n, c, h, w = self.x.shape
        f = self.upscale_factor.val
        ret_shape = (n, c // (f * f), h * f, w * f)
        return types.tensor(x_type, ret_shape)


@register_op(doc_str="")
class sliding_windows(Operation):
    """
    Return a tensor containing all windows of ``size``, separated by stride along the
    given ``axis``.
    
    Parameters
    ----------
    x: tensor<[\*d0, d_axis, *dn], T>
        * Input tensor.
    
    axis: const<i32>
        * Axis to perform the operation.
    
    size: const<i32>
        * Number of elements in the sliding window.
    
    stride: const<i32> Optional
        * Default to ``1``.
        * The stride of the input elements in the sliding window.
    
    Returns
    -------
    tensor<[\*d0, d_axis - size // stride + 1, size, \*dn], T>
        * The output will be a tensor of rank ``N+1`` where ``N`` is the input tensor
          rank.
    
    Attributes
    ----------
    T: fp32
    """
    
    input_spec = InputSpec(
        x=TensorInputType(),
        axis=IntInputType(const=True),
        size=IntInputType(const=True),
        stride=IntInputType(const=True, optional=True),
    )

    def default_inputs(self):
        return DefaultInputs(stride=1)

    def __init__(self, **kwargs):
        super(sliding_windows, self).__init__(**kwargs)

    def type_inference(self):
        x_shape = self.x.shape
        axis = self.axis.val
        size = self.size.val
        stride = self.stride.val
        ret_shape = list(x_shape)
        ret_shape[axis] = (x_shape[axis] - size) // stride + 1
        pos_axis = axis if axis >= 0 else axis + self.x.rank
        ret_shape.insert(pos_axis + 1, size)
        return types.tensor(self.x.dtype, tuple(ret_shape))



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import numpy as np

from coremltools.converters.mil.mil.ops.defs._utils import (
    aggregated_pad,
    effective_kernel,
    spatial_dimensions_out_shape,
)


class TestDilation:
    def test_kernel_and_dilations_not_same_size(self):
        np.testing.assert_raises_regex(
            ValueError,
            "kernel_shape.*dilations.*length",
            effective_kernel,
            kernel_shape=(1, 2, 3),
            dilations=(1, 2),
        )

    def test_effective_kernel_dilation_1(self):
        actual = effective_kernel(kernel_shape=(1, 2, 3), dilations=(1, 1, 1))

        expected = [1, 2, 3]
        np.testing.assert_equal(actual, expected)

    def test_effective_kernel_dilation_2(self):
        actual = effective_kernel(kernel_shape=(1, 2, 3), dilations=(2, 2, 2))

        expected = [1, 3, 5]
        np.testing.assert_equal(actual, expected)

    def test_effective_kernel_dilation_3(self):
        actual = effective_kernel(kernel_shape=(1, 2, 3), dilations=(3, 3, 3))

        expected = [1, 4, 7]
        np.testing.assert_equal(actual, expected)


class TestAggregatePadding:
    def test_invalid_pad_type(self):
        np.testing.assert_raises_regex(
            ValueError,
            "Invalid padding pad_type",
            aggregated_pad,
            pad_type="bananas",
            kernel_shape=(1, 2, 3),
        )

    def test_dilations_rank_different_from_input_rank(self):
        np.testing.assert_raises_regex(
            ValueError,
            "dilations must have same length as kernel_shape",
            aggregated_pad,
            pad_type="valid",  # doesn't matter
            kernel_shape=(1, 2, 3),
            dilations=(4, 5),
        )

    def test_custom_pad(self):
        actual = aggregated_pad(
            pad_type="custom", kernel_shape=(1, 2, 3), custom_pad=(7, 8, 9, 10, 11, 12)
        )

        expected = [7 + 8, 9 + 10, 11 + 12]
        np.testing.assert_equal(actual, expected)

    def test_custom_pad_none(self):
        np.testing.assert_raises_regex(
            ValueError,
            "Invalid custom_pad",
            aggregated_pad,
            pad_type="custom",
            kernel_shape=(1, 2, 3),  # doesn't matter
            custom_pad=None,
        )

    def test_custom_pad_invalid(self):
        np.testing.assert_raises_regex(
            ValueError,
            "Invalid custom_pad",
            aggregated_pad,
            pad_type="custom",
            kernel_shape=(1, 2, 3),  # doesn't matter
            custom_pad=(7, 8, 9, 10),  # too few elements
        )

    def test_valid_pad(self):
        actual = aggregated_pad(pad_type="valid", kernel_shape=(1, 2, 3),)

        expected = [0, 0, 0]
        np.testing.assert_equal(actual, expected)

    def test_valid_pad_4d(self):
        actual = aggregated_pad(pad_type="valid", kernel_shape=(1, 2, 3, 4),)

        expected = [0, 0, 0, 0]
        np.testing.assert_equal(actual, expected)

    def test_valid_pad_2d(self):
        actual = aggregated_pad(pad_type="valid", kernel_shape=(1, 2),)

        expected = [0, 0]
        np.testing.assert_equal(actual, expected)

    def test_valid_pad_1d(self):
        actual = aggregated_pad(pad_type="valid", kernel_shape=[4])

        expected = [0]
        np.testing.assert_equal(actual, expected)

    def test_same_padding_no_dilation(self):
        actual = aggregated_pad(
            pad_type="same",
            input_shape=(5, 6, 7),
            kernel_shape=(2, 2, 2),
            strides=(1, 2, 2),
        )

        expected = [1, 0, 1]
        np.testing.assert_equal(actual, expected)

    def test_same_padding_dilation_with_dilation(self):
        actual = aggregated_pad(
            pad_type="same",
            input_shape=(19, 20, 21),
            kernel_shape=(2, 2, 2),
            strides=(1, 2, 2),
            dilations=(5, 6, 7),
        )

        expected = [5, 5, 7]
        np.testing.assert_equal(actual, expected)

    def test_same_padding_stride_same_as_input(self):
        actual = aggregated_pad(
            pad_type="same", input_shape=(5, 5), kernel_shape=(3, 3), strides=(5, 5),
        )

        expected = [0, 0]
        np.testing.assert_equal(actual, expected)

    def test_same_padding_stride_larger_than_kernel_but_less_than_input(self):
        actual = aggregated_pad(
            pad_type="same", input_shape=(5, 5), kernel_shape=(3, 3), strides=(4, 4),
        )

        expected = [2, 2]
        np.testing.assert_equal(actual, expected)

    def test_same_padding_none_input_shape(self):
        np.testing.assert_raises_regex(
            ValueError,
            "input_shape.*None",
            aggregated_pad,
            pad_type="same",
            kernel_shape=(1, 2, 3),
            strides=(1, 2, 3),
        )

    def test_same_padding_input_shape_wrong_size(self):
        np.testing.assert_raises_regex(
            ValueError,
            "input_shape.*same length",
            aggregated_pad,
            pad_type="same",
            kernel_shape=(1, 2, 3),
            input_shape=(1, 2),
            strides=(1, 2, 3),
        )

    def test_same_padding_none_strides(self):
        np.testing.assert_raises_regex(
            ValueError,
            "strides.*None",
            aggregated_pad,
            pad_type="same",
            kernel_shape=(1, 2, 3),
            input_shape=(1, 2, 3),
        )

    def test_same_padding_strides_wrong_size(self):
        np.testing.assert_raises_regex(
            ValueError,
            "strides.*same length",
            aggregated_pad,
            pad_type="same",
            kernel_shape=(1, 2, 3),
            input_shape=(1, 2, 3),
            strides=(1, 2),
        )


class TestOutputShape:
    def test_custom_padding_shape(self):
        actual = spatial_dimensions_out_shape(
            pad_type="custom",
            input_shape=(3, 3, 3),
            kernel_shape=(2, 2, 2),
            strides=(2, 2, 2),
            custom_pad=(2, 0, 1, 2, 2, 3),
        )

        expected = [2, 3, 4]
        np.testing.assert_equal(actual, expected)

    def test_valid_padding_shape(self):
        actual = spatial_dimensions_out_shape(
            pad_type="valid", input_shape=(7, 7), kernel_shape=(3, 3), strides=(1, 1)
        )

        expected = [5, 5]
        np.testing.assert_equal(actual, expected)

    def test_valid_padding_shape_dilation_2(self):
        actual = spatial_dimensions_out_shape(
            pad_type="valid",
            input_shape=(7, 7),
            kernel_shape=(3, 3),
            strides=(1, 1),
            dilations=(2, 2),
        )

        expected = [3, 3]
        np.testing.assert_equal(actual, expected)

    def test_valid_padding_shape_with_stride_2(self):
        actual = spatial_dimensions_out_shape(
            pad_type="valid", input_shape=(7, 7), kernel_shape=(3, 3), strides=(2, 2)
        )

        expected = [3, 3]
        np.testing.assert_equal(actual, expected)

    def test_same_padding_shape(self):
        actual = spatial_dimensions_out_shape(
            pad_type="same", input_shape=(6, 6), kernel_shape=(2, 2), strides=(2, 2)
        )

        expected = [3, 3]
        np.testing.assert_equal(actual, expected)

    def test_same_padding_shape_stride_2_input_not_multiple_of_kernel(self):
        actual = spatial_dimensions_out_shape(
            pad_type="same", input_shape=(5, 5), kernel_shape=(2, 2), strides=(2, 2)
        )

        expected = [3, 3]
        np.testing.assert_equal(actual, expected)

    def test_same_padding_shape_dilation_2(self):
        actual = spatial_dimensions_out_shape(
            pad_type="same",
            input_shape=(5, 5),
            kernel_shape=(2, 2),
            strides=(1, 1),
            dilations=(2, 2),
        )

        expected = [5, 5]
        np.testing.assert_equal(actual, expected)
